/*
 * aseread.c -- Interface to the ASEDrive reader
 *
 * COPYRIGHT:
 *	Copyright 1997 Wolf Geldmacher, Comunicon AG, CH-9052 Niederteufen.
 *	All rights reserved. See file COPYING.
 *
 */

#include "aseread.h"
RCSID "$Id: aseread.c,v 1.3 1998/04/18 20:27:39 wjg Exp $";

#if INCLUDE_ASE != 0

#include <assert.h>
#include "ase.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Create Prototypes for ANSI-C
 */
#ifdef __STDC__
#define P(x) x
#else
#define P(x) ()
#endif

/*
 * Declaration of local functions
 */
static int ase_init P((
	card_reader *reader,
	int port,
	int timeout));
static int ase_on P((
	card_reader *reader,
	unsigned char *atr,
	unsigned int *atrlenp));
static int ase_send P((
	card_reader *reader,
	unsigned char cla,
	unsigned char ins,
	unsigned char p1,
	unsigned char p2,
	unsigned short len,
	unsigned char *data,
	unsigned char *ackp,
	unsigned char *sw1p,
	unsigned char *sw2p));
static int ase_recv P((
	card_reader *reader,
	unsigned char cla,
	unsigned char ins,
	unsigned char p1,
	unsigned char p2,
	unsigned short len,
	unsigned char *data,
	unsigned char *ackp,
	unsigned char *sw1p,
	unsigned char *sw2p));
static int ase_off P((
	card_reader *reader));
static char *ase_msg P((
	card_reader *reader));
static char *ase_version P((
	card_reader *reader));

#undef P

#ifdef __cplusplus
}
#endif

/*
 * Exported interface to ASEDrive
 */
card_reader AseReader = {
	"ASEDrive",		/* reader name			*/
	ase_init,		/* reader initialize command	*/
	ase_on,			/* reader on/reset command	*/
	ase_send,		/* reader send command		*/
	ase_recv,		/* reader recv command		*/
	ase_off,		/* reader off command		*/
	ase_msg,		/* reader status message	*/
	ase_version,		/* return version information	*/
	0			/* reader status		*/
};

static int
ase_init(reader, port, timeout)
card_reader *reader;
int port;
int timeout;
{
	reader->status = ASE_DRVInit((short)port, (short)timeout);
	return reader->status == 0;
}

/*
 * ase_on -- initialize and reset card
 *	This function is somewhat complicated because we must interpret
 *	the ATR returned by the drive.
 *
 *	Unfortuately the ASE interface does not allow us to specify a buffer
 *	size bigger than the actual resetLen of the card!
 *
 *	Reset response length varies from minimum 2 bytes to maximum
 *	the number of bytes spcified in the second byte plus additional
 *	TDi bytes (max 33). The way we handle this with the ASE interface
 *	is to first get the initial two bytes (gives us the number of
 *	Tx1 chars plus the number of historical bytes). Then, if the TDi
 *	byte has additional information, we get that much more. This will
 *	lead to several reset instructions issued to the card - which
 *	violates the ISO7816/3 Amendment 2 specification, if the TA2 byte
 *	is present.
 */
#define N	4			/* Invalid Txi index		*/
static int
ase_on(reader, atr, atrlenp)
card_reader *reader;
unsigned char *atr;
unsigned int *atrlenp;
{
	static const char bits[16] =	/* # of bits in a nibble	*/
		{ 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4 };
	static const char tci[16] =	/* index of TCi byte		*/
		{ N, N, N, N, 0, 1, 1, 2, N, N, N, N, 0, 1, 1, 2 };
	static const char tdi[16] =	/* index of TDi byte		*/
		{ N, N, N, N, N, N, N, N, 0, 1, 1, 2, 1, 2, 2, 3 };
	short guardLen = 0;		/* default guard time value	*/
	short resetLen = 2;		/* length of ATR response	*/
	int histChars = 0;		/* number of historical chars	*/
	int txChars = 0;		/* number of Tx characters	*/
	unsigned txBits;		/* Bits in T0 and TDi		*/
	int haveTck = FALSE;		/* Set if we have a TCK byte	*/
	int i;

	assert(atr && atrlenp && *atrlenp >= 2);
	memset(atr, '\0', *atrlenp);

	/*
	 * Note that any ISO7816 complying card must answer with at least
	 * two bytes: TS and T0.
	 */
	if ((reader->status = ASE_7816CardOn(guardLen, resetLen, atr)))
		return 0;

	/*
	 * Interpret T0: count of histChars and Tx1 bytes
	 */
	histChars = atr[1] & 0x0F;
	resetLen += histChars;

	/*
	 * Use information in high nibble of T0/TDi to get all of the
	 * information the card provides.
	 */
	i = 1;
	txBits = ((atr[i] & 0xF0) >> 4);
	resetLen += bits[txBits];
	txChars += bits[txBits];
	while (tdi[txBits] != N) {
		/*
		 * If we have a TDi byte, we could also have a TCK byte after
		 * all of the current ATR. The presence of a TCK byte is
		 * determined by ((TDi & 0x0F) != 0), i.e. any protocol
		 * accepted other than T0.
		 */
		if (!haveTck && (atr[i + tdi[txBits]] & 0xF)) {
			haveTck = TRUE;
			++resetLen;
		}

		/*
		 * Check for buffer overflow.
		 */
		assert((unsigned)resetLen <= *atrlenp);

		if ((reader->status = ASE_7816CardOn(guardLen, resetLen, atr)))
			return 0;

		i += bits[(atr[i] & 0xF0) >> 4];
		txBits = ((atr[i] & 0xF0) >> 4);
		resetLen += bits[txBits];
		txChars += bits[txBits];
	}

	/*
	 * Perform reset to get all of the data
	 */
	if ((reader->status = ASE_7816CardOn(guardLen, resetLen, atr)))
		return 0;

	/*
	 * Evaluate TC1, if present, to set guardLen to correct value.
	 *	0..254 - extra guard time required (in ETU's)
	 *	255    - guard time reduced from 12 ETU's to 11 ETU's
	 */
	txBits = ((atr[1] & 0xF0) >> 4);
	if (tci[txBits] != N && (guardLen = atr[2 + tci[txBits]])) {
		/*
		 * Perform reset so correct guardLen is provided to drive.
		 */
		if ((reader->status = ASE_7816CardOn(guardLen, resetLen, atr)))
			return 0;
	}

	*atrlenp = resetLen;
	return 1;
}
#undef N

static int
ase_send(reader, cla, ins, p1, p2, len, data, ackp, sw1p, sw2p)
card_reader *reader;	/* Pointer to reader information	*/
unsigned char cla;	/* I: Class byte			*/
unsigned char ins;	/* I: Instruction byte			*/
unsigned char p1;	/* I: P1 byte				*/
unsigned char p2;	/* I: P2 byte				*/
unsigned short len;	/* I: LEN of data			*/
unsigned char *data;	/* I: data to send			*/
unsigned char *ackp;	/* O: Acknowledge byte 			*/
unsigned char *sw1p;	/* O: Contains SW1 for last call	*/
unsigned char *sw2p;	/* O: Contains SW2 for last call	*/
{
	assert(reader);

	reader->status = ASE_7816Send((short)cla, (short)ins, (short)p1,
				(short)p2, (short)len, data, ackp, sw1p, sw2p);

	return (reader->status == 0 || reader->status == 0x9000);
}

static int
ase_recv(reader, cla, ins, p1, p2, len, data, ackp, sw1p, sw2p)
card_reader *reader;	/* Pointer to reader information	*/
unsigned char cla;	/* I: Class byte			*/
unsigned char ins;	/* I: Instruction byte			*/
unsigned char p1;	/* I: P1 byte				*/
unsigned char p2;	/* I: P2 byte				*/
unsigned short len;	/* I: LEN of data			*/
unsigned char *data;	/* O: data received			*/
unsigned char *ackp;	/* O: Acknowledge byte 			*/
unsigned char *sw1p;	/* O: Contains SW1 for last call	*/
unsigned char *sw2p;	/* O: Contains SW2 for last call	*/
{
	assert(reader);

	reader->status = ASE_7816Receive((short)cla, (short)ins, (short)p1,
				(short)p2, (short)len, data, ackp, sw1p, sw2p);

	return (reader->status == 0 || reader->status == 0x9000);
}

static int
ase_off(reader)
card_reader *reader;
{
	assert(reader);

	reader->status = ASE_7816CardOff();

	return reader->status == 0;
}

/*
 * Function to get ASE drive and software information
 */
static char *
ase_version(reader)
card_reader *reader;
{
	static char tbuf[80];
	char vbuf[4];

	if ((reader->status = ASE_GetVersion(vbuf, sizeof(vbuf))) != 0)
		return 0;
	sprintf(tbuf, "ASEDrive V%d.%d, ASESoft V%d.%d", vbuf[0] & 0xFF,
		vbuf[1] & 0xFF, vbuf[2] & 0xFF, vbuf[3] & 0xFF);
	return tbuf;
}

/*
 * Function to decode an ASE drive status code
 */
static char *
ase_msg(reader)
card_reader *reader;
{
	static char tbuf[80];

	assert(reader);

	switch (reader->status) {
	case   0: return "No error";
	case   1: return "ASEDrive communication error";
	case   4: return "ASEDrive not found";
	case  70: return "ASEDrive busy";
	case  71: return "PC detected a checksum error";
	case  79: return "Command not recognized";
	case  81: return "ASEDrive detected an error in the communication";
	case  82: return "ASEDrive detected a checksum error";
	case  83: return "The data length parameter is too big for the ASEDrive";
	case  84: return "Card power on error";
	case  85: return "Unknown card reset response";
	case  86: return "Command aborted by the card";
	case  87: return "Unsupported smartcard";
	case  88: return "ASEDrive battery is low";
	case  89: return "Card power off error";
	case  90: return "Command aborted";
	case  91: return "No card response";
	case  92: return "Smartcard not defined";
	case  93: return "Smartcard error";
	case  94: return "Command error";
	case  95: return "Unknown ASEDrive command";
	case 128: return "Cannot open Windows NT device driver";
	case 129: return "Comand aborted by device driver";
	default:  sprintf(tbuf, "Unknown error code 0x%04x (%d)",
			reader->status, reader->status);
		  return tbuf;
	}
}

#endif
