/*
 * TODO:
 *	- sync of dp->??? with TCL variables
 *	- command with long arguments
 *	- card is xxx: locate file in path
 *	- Card_Init: locate module specific startup file.
 *
 * DESCRIPTION:
 *	Implementation of TCL interface to ISO chipcards.
 *
 * COPYRIGHT:
 *	Copyright 1997 Wolf Geldmacher, Comunicon AG, CH-9052 Niederteufen.
 *	All rights reserved. See file COPYING.
 *
 * This implementation provides the following new TCL commands:
 *
 *	card init [-reader type] [-port port] [-timeout timeout]
 *		This command (re-)initializes the reader at the port specified
 *		and optionally sets the timeout of the drive to the number of
 *		seconds specified. A card must be inserted in the reader but
 *		no command is actually sent to the card.
 *
 *		The reader option specifies the type of card reader connected.
 *		If not given all known reader types are tried in turn.
 *
 *		If the port is not specified or is 0 the init command tries
 *		to automatically locate the first reader (of the given type)
 *		connected. The following port numbers are valid (and are
 *		searched in this order to locate the reader):
 *			1..4	first to fourth serial port
 *			5..7	first to third parallel port
 *			0		auto detection
 *
 *		If the timeout is not specified it defaults to 5 seconds.
 *
 *		Returns an empty string on success. On error a descriptive
 *		error message is output and the command fails.
 *
 *	card reset
 *		This command applies power to the chip card and gets the ATR.
 *
 *		Returns the hexlist ATR of the card and sets the verbose status
 *		to a TCL list that decodes the ATR according to ISO7816. On
 *		error a descriptive error message is output and the command
 *		fails.
 *
 *	card status -verbose
 *	card status -add code message
 *	card status -delete code
 *	card status -list
 *	card status -reader
 *	card status -version
 *		This command retrieves the status of the last card command
 *		issued. If the verbose option is specified a full text version
 *		of the status information is returned, if available.
 *
 *		Returns the value of SW1, SW2 and ACK. A list of three hex
 *		values.
 *
 *		The second and third form of the statement add respectivly
 *		delete verbose messages for status codes. The status code
 *		may contain one or more consecutive '?' as wildcards. The
 *		value represented by those wildcards is available as one
 *		integer argument (printf style format) to the message.
 *
 *		The list option creates a list of code/message
 *		pairs that are known to the status command.
 *
 *		The -reader option returns the status of the reader.
 *
 *		The -version option returns HW and SW level of the initialized
 *		card reader.
 *
 *	card command CLA INS P1 P2 [-send data_list] [-receive length]
 *		This command sends and/or receives data from the chipcard. It
 *		handles all the different formats (1..4) such a message can
 *		take:
 *			If data_list is an empty list no data is sent, otherwise
 *			the number of bytes in the argument hexlist are
 *			counted and that many bytes are sent (count is passed
 *			as P3 as specified by ISO). If -receive length is
 *			specified that many bytes (0 = all) are waited for and
 *			returned.
 *
 *		Either -send or -receive or both must be specified. The order
 *		of specification is free.
 *
 *		Returns the data generated by the command.
 *
 *	card "command" args
 *		This command calls the card sepcific command of a card announced
 *		by the 'card is' command.
 *
 *	card off
 *		This command turns the card's power off. It also terminates
 *		the connection to the card reader so you have to reissue both
 *		card init and card on before you can continue to work with
 *		another smartcard.
 *		Returns nothing.
 *
 *	card is [name]
 *		This command shows or sets the card type.
 *
 *		If called with three arguments it sources the card specific tcl
 *		module. It also adds the card specific commands to the card
 *		meta command. The card specific module is searched for in the
 *		tcl_library_path. A card specific module contains an array 
 *		ariable with the name of the card module and card specific
 *		commands prefixed by the card name, e.g.: cc4(rootkey) or
 *		cc4:select.
 *
 *		If called without the name argument returns the current name
 *		of the card (same as $card(CARD)).
 *
 * In addition to these new commands the global array 'card' is defined
 * with the following members:
 *	card(VERSION)	-- version of the card library
 *	card(READER)	-- reader currently in use
 *	card(TS)	-- TS byte of ATR as hexadecimal number
 *	card(T0)	-- T0 byte of ATR as hexadecimal number
 *	card(Txi)	-- Txi bytes of ATR as hexadecimal numbers
 *	card(HIST)	-- The historical bytes as hexlist
 *	card(ATR)	-- The whole ATR as hexlist
 *	card(ATR_VERBOSE) -- Verbose version of the ATR
 *	card(TCK)	-- Value of the check byte, if present
 *	card(SW1)	-- SW1 byte of last command
 *	card(SW2)	-- SW2 byte of last command
 *	card(ACK)	-- ACK byte of last command
 *	card(STATUS)	-- verbose status message (SW1/SW2 decoded)
 *	card(COMMAND)	-- last command issued to the card
 *	card(CARD)	-- card type
 */

#include "card.h"
#include "hex.h"
#include "status.h"
#include "tcl.h"

RCSID "$Id: card.c,v 1.5 2000/10/12 22:18:20 wjg Exp $";

#if INCLUDE_ASE != 0
#include "aseread.h"
#endif
#if INCLUDE_GPR != 0
#include "gprread.h"
#endif
#if INCLUDE_REFLEX != 0
#include "rflxread.h"
#endif

/*
 * Macros for fast access
 */
#define IS_CMD(x, y)	(strncmp(x, y, strlen(x)) == 0)

/*
 * Array of known readers
 */
static card_reader *Readers[] = {
#if INCLUDE_ASE != 0
	&AseReader,		/* ASE Drive				*/
#endif
#if INCLUDE_GPR != 0
	&GprReader,		/* GPR400 Drive				*/
#endif
#if INCLUDE_REFLEX != 0
	&ReflexReader,		/* Schlumberger Reflex 60 Drive		*/
#endif
	0
};
static char crd[] = "card";	/* to save room */

/*
 * ANSI-C prototypes
 */
static void setStatus
	_ANSI_ARGS_((Tcl_Interp *interp, card_data *dp));

static void cardfree
	_ANSI_ARGS_((ClientData clientData));

static int card
	_ANSI_ARGS_((ClientData clientData, Tcl_Interp *interp, int argc,
		char **argv));

/*
 * getReader -- helper function for card init
 */
static int
getReader(interp, arg, rpp)
Tcl_Interp *interp;
char *arg;
card_reader **rpp;
{
	int i;
	card_reader *rp;
	int length = strlen(arg);

	for (i = 0; (rp = Readers[i]); ++i) {
		if (strncmp(arg, rp->name, length) == 0) {
			*rpp = rp;
			return TCL_OK;
		}
	}

	Tcl_AppendResult(interp, "card init: reader \"", arg,
		"\" is not known", NULL);
	return TCL_ERROR;
}

/*
 * getTimeout -- helper function for card init
 */
static int
getTimeout(interp, arg, top)
Tcl_Interp *interp;
char *arg;
int *top;
{
	if (*top != -1) {
		Tcl_AppendResult(interp,
			"card init: duplicate -timeout option",
			NULL);
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, arg, top) != TCL_OK) {
		Tcl_AppendResult(interp,
			"card init: -timeout needs numeric argument",
			NULL);
		return TCL_ERROR;
	}
	if (*top <= 0 || *top > 500) {
		Tcl_AppendResult(interp,
			"card init: -timeout value must be in the range 1..500",
			NULL);
		return TCL_ERROR;
	}
	return TCL_OK;
}

/*
 * getPort -- helper function for card init
 */
static int
getPort(interp, arg, portp)
Tcl_Interp *interp;
char *arg;
int *portp;
{
	if (*portp != -1) {
		Tcl_AppendResult(interp,
			"card init: duplicate -port option",
			NULL);
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, arg, portp) != TCL_OK) {
		Tcl_AppendResult(interp,
			"card init: -port needs numeric argument",
			NULL);
		return TCL_ERROR;
	}
	if (*portp < 0 || *portp > 7) {
		Tcl_AppendResult(interp,
			"card init: -port value must be in the range 0..7",
			NULL);
		return TCL_ERROR;
	}
	return TCL_OK;
}

/*
 * init -- select and initialize a card reader
 */
static int
init(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	card_data *dp = (card_data *)clientData;
	card_reader *rp = dp->reader;
	char *msg = 0;
	int port = -1;
	int timeout = -1;
	int i;

	/*
	 * Handle options
	 */
	for (i = 2; i < argc; ++i) {
		int length = strlen(argv[i]);
		if (strncmp(argv[i], "-reader", length) == 0) {
			if (++i >= argc) {
				msg = "card init: -reader needs an argument";
				break;
			}
			if (getReader(interp, argv[i], &rp) != TCL_OK)
				return TCL_ERROR;
		} else if (strncmp(argv[i], "-timeout", length) == 0) {
			if (++i >= argc) {
				msg = "card init: -timeout needs an argument";
				break;
			}
			if (getTimeout(interp, argv[i], &timeout) != TCL_OK)
				return TCL_ERROR;
		} else if (strncmp(argv[i], "-port", length) == 0) {
			if (++i >= argc) {
				msg = "card init: -port needs an argument";
				break;
			}
			if (getPort(interp, argv[i], &port) != TCL_OK)
				return TCL_ERROR;
		} else {
			msg = "usage: card init [-reader reader] [-port port] [-timeout timeout]";
			break;
		}
	}

	if (msg) {
		Tcl_AppendResult(interp, msg, NULL);
		return TCL_ERROR;
	}

	/*
	 * Set default values for options not specified
	 */
	if (port < 0)
		port = 0;

	if (timeout < 0)
		timeout = 5;

	/*
	 * If no reader was specified try to find reader by calling
	 * initialization for every reader in turn.
	 */
	if (rp == 0) {
		for (i = 0; (rp = Readers[i]); ++i)
			if (rp->init(rp, port, timeout) == TRUE)
				break;

		if (rp == 0) {
			Tcl_AppendResult(interp,
				"card init: no card reader found",
				NULL);
			return TCL_ERROR;
		}
		dp->reader = rp;
		return TCL_OK;
	}

	/*
	 * Initialize the reader
	 */
	if (!rp->init(rp, port, timeout)) {
		Tcl_AppendResult(interp, "card init: ", rp->stat(rp), NULL);
		return TCL_ERROR;
	}

	return TCL_OK;
}

/*
 * reset -- reset a card an get its ATR
 */
#define N	4			/* Invalid Txi index		*/
static int
reset(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	static const char bits[16] =	/* # of bits in a nibble	*/
		{ 0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4 };
	static const char tai[16] =	/* index of TAi byte,		*/
		{ N, 0, N, 0, N, 0, N, 0, N, 0, N, 0, N, 0, N, 0 };
	static const char tbi[16] =	/* index of TBi byte		*/
		{ N, N, 0, 1, N, N, 0, 1, N, N, 0, 1, N, N, 0, 1 };
	static const char tci[16] =	/* index of TCi byte		*/
		{ N, N, N, N, 0, 1, 1, 2, N, N, N, N, 0, 1, 1, 2 };
	static const char tdi[16] =	/* index of TDi byte		*/
		{ N, N, N, N, N, N, N, N, 0, 1, 1, 2, 1, 2, 2, 3 };
	card_data *dp = (card_data *)clientData;
	card_reader *rp = dp->reader;
	unsigned char resetBuf[256];	/* ISO: max of TS + 32 bytes	*/
	int resetLen = sizeof(resetBuf);/* max. length of ATR response	*/
	char tbuf[256];			/* temp buffer for msgs		*/
	int len;			/* current length of tbuf data	*/
	unsigned char val;		/* value of Tx byte		*/
	int histChars;			/* number of historical chars	*/
	int txChars = 0;		/* number of Tx characters	*/
	unsigned txBits;		/* Bits in T0 and TDi		*/
	int haveTck = FALSE;		/* Set if we have a TCK byte	*/
	int i,j; 
	char *tmp;

	/*
	 * Check for initialized reader
	 */
	if (rp == 0) {
		Tcl_AppendResult(interp, "card reset: reader not initialized",
			NULL);
		return TCL_ERROR;
	}

	/*
	 * Check argument list
	 */
	if (argc != 2) {
		Tcl_AppendResult(interp, "card reset: too many arguments",
			NULL);
		return TCL_ERROR;
	}

	/*
	 * Unset array variable. This leaves us with a clean start.
	 */
	tmp = Tcl_GetVar2(interp, crd, "CARD", TCL_GLOBAL_ONLY);
	if (tmp)
		tmp = strdup(tmp);
	Tcl_UnsetVar(interp, crd, TCL_GLOBAL_ONLY);
	Tcl_SetVar2(interp, crd, "VERSION", CARD_VERSION,
		TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
	if (tmp) {
		Tcl_SetVar2(interp, crd, "CARD", tmp, TCL_GLOBAL_ONLY);
		free(tmp);
	}

	/*
	 * Reset the card.
	 */
	if (rp->on(rp, resetBuf, &resetLen) != 1) {
		Tcl_AppendResult(interp, "card reset: ", rp->stat(rp), NULL);
		return TCL_ERROR;
	}

	/*
	 * Interpret the TS and T0 character
	 */
	if (resetBuf[0] == 0x3F)
		Tcl_SetVar2(interp, crd, "ATR_VERBOSE",
			"TS 0x3F: inverse convention",
			TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
	else if (resetBuf[0] == 0x3B)
		Tcl_SetVar2(interp, crd, "ATR_VERBOSE",
			"TS 0x3B: direct convention",
			TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
	else {
		Tcl_AppendResult(interp, "card reset: Invalid TS", NULL);
		return TCL_ERROR;
	}
	sprintf(tbuf, "0x%02x", resetBuf[0]);
	Tcl_SetVar2(interp, crd, "TS", tbuf,
		TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

	/*
	 * T0 is interpreted the following way:
	 *	High nibble:
	 *		b8 == 1 - TD1 follows
	 *		b7 == 1 - TC1 follows
	 *		b6 == 1 - TB1 follows
	 *		b5 == 1 - TA1 follows
	 * Low nibble:
	 *		0..15   - number of historical characters that follow
	 */
	val = resetBuf[1];
	sprintf(tbuf, "0x%02x", val);
	Tcl_SetVar2(interp, crd, "T0", tbuf,
		TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
	histChars = resetBuf[1] & 0x0F;
	txChars += bits[(val & 0xF0) >> 4];

	len = sprintf(tbuf, "T0 0x%02x:", val);
	if ((val & 0xF0) == 0)
		len += sprintf(&tbuf[len], " No Tx1 bytes,");
	else {
		if (val & 0x10)
			len += sprintf(&tbuf[len], " TA1,");
		if (val & 0x20)
			len += sprintf(&tbuf[len], " TB1,");
		if (val & 0x40)
			len += sprintf(&tbuf[len], " TC1,");
		if (val & 0x80)
			len += sprintf(&tbuf[len], " TD1,");
	}
	len += sprintf(&tbuf[len], " and %d historical characters follow",
			histChars);
	Tcl_SetVar2(interp, crd, "ATR_VERBOSE", tbuf,
		TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

	/*
	 * Now look at what we've got
	 */
	txBits = ((resetBuf[1] & 0xF0) >> 4);
	i = 2;			/* base offset into resetBuf		*/
	j = 1;			/* x in Tx				*/
	do {
		if (tai[txBits] != N) {		/* TAi follows		*/
			val = resetBuf[i + tai[txBits]];

			len = sprintf(tbuf, "TA%d", j);
			sprintf(&tbuf[len + 1], "0x%02x", val);
			Tcl_SetVar2(interp, crd, tbuf, &tbuf[len+1],
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
			switch (j) {
			case 1:
				/*
				 * TA1 is interpreted the following way:
				 *	High nibble: clock conversion factor F
				 *	Low nibble: Bit rate adjustment factor D
				 *
				 * initial clock etu = 372/fi s, where fi is the initial CLK
				 *	frequency of 1..5 MHz
				 * work etu = 1/D * F/fs s, where fs is the subsequent CLK
				 *	frequency (i.e. the frequency applied after reset) of 1..max MHz
				 */
				len = sprintf(tbuf, "TA1 0x%02x: clock conversion F is", val);
				switch (val & 0xF0) {
				case 0x00:	len += sprintf(&tbuf[len], " (internal clock)"); break;
				case 0x10:	len += sprintf(&tbuf[len], " 372 (max. 5 MHz)"); break;
				case 0x20:	len += sprintf(&tbuf[len], " 558 (max. 6 MHz)"); break;
				case 0x30:	len += sprintf(&tbuf[len], " 744 (max. 8 MHz)"); break;
				case 0x40:	len += sprintf(&tbuf[len], " 1116 (max. 12 MHz)"); break;
				case 0x50:	len += sprintf(&tbuf[len], " 1488 (max. 16 MHz)"); break;
				case 0x60:	len += sprintf(&tbuf[len], " 1860 (max. 20 MHz)"); break;
				case 0x90:	len += sprintf(&tbuf[len], " 512 (max. 5 MHz)"); break;
				case 0xA0:	len += sprintf(&tbuf[len], " 768 (max. 7.5 MHz)"); break;
				case 0xB0:	len += sprintf(&tbuf[len], " 1024 (max. 10 MHz)"); break;
				case 0xC0:	len += sprintf(&tbuf[len], " 1536 (max. 15 MHz)"); break;
				case 0xD0:	len += sprintf(&tbuf[len], " 2048 (max. 20 MHz)"); break;
				default:	len += sprintf(&tbuf[len], " RFU"); break;
				}

				len += sprintf(&tbuf[len], ", bit rate conversion D is");
				switch (val & 0x0F) {
				case 0x01:	len += sprintf(&tbuf[len], " 1"); break;
				case 0x02:	len += sprintf(&tbuf[len], " 2"); break;
				case 0x03:	len += sprintf(&tbuf[len], " 4"); break;
				case 0x04:	len += sprintf(&tbuf[len], " 8"); break;
				case 0x05:	len += sprintf(&tbuf[len], " 16"); break;
				case 0x06:	len += sprintf(&tbuf[len], " 32"); break;
				case 0x08:	len += sprintf(&tbuf[len], " 12"); break;
				case 0x09:	len += sprintf(&tbuf[len], " 20"); break;
				case 0x0A:	len += sprintf(&tbuf[len], " 1/2"); break;
				case 0x0B:	len += sprintf(&tbuf[len], " 1/4"); break;
				case 0x0C:	len += sprintf(&tbuf[len], " 1/8"); break;
				case 0x0D:	len += sprintf(&tbuf[len], " 1/16"); break;
				case 0x0E:	len += sprintf(&tbuf[len], " 1/32"); break;
				case 0x0F:	len += sprintf(&tbuf[len], " 1/64"); break;
				default:	len += sprintf(&tbuf[len], " RFU"); break;
				}
				break;

			case 2:
				/*
				 * The TA2 byte indicates a specific mode. Note that
				 * according to ISO7816/3 Amd2 the card may switch between
				 * negotiable mode and specific mode given more than
				 * one reset. It also states that we should *not* issue
				 * additional resets until we have completely received the
				 * first ATR - which is not possible with the ASE interface!
				 */
				len = sprintf(tbuf, "TA2 0x%02x: specific mode", val);
				if (val & 0x80)
					len += sprintf(&tbuf[len], " not capable to change mode");
				else
					len += sprintf(&tbuf[len], " capable to change mode");
				if (val & 0x10)
					len += sprintf(&tbuf[len], ", implicitly defined");
				else
					len += sprintf(&tbuf[len], ", defined by interface bytes");

				switch (val & 0xF) {
				case 0:	len += sprintf(&tbuf[len], " for T0 protocol (async half duplex char)"); break;
				case 1:	len += sprintf(&tbuf[len], " for T1 protocol (async half duplex block)"); break;
				case 2:	len += sprintf(&tbuf[len], " for T2 protocol RFU (full duplex)"); break;
				case 3:	len += sprintf(&tbuf[len], " for T3 protocol RFU (full duplex)"); break;
				case 4:	len += sprintf(&tbuf[len], " for T4 protocol (enhanced async half duplex char)"); break;
				case 14:len += sprintf(&tbuf[len], " for T14 protocol (Non ISO)"); break;
				default:len += sprintf(&tbuf[len], " for T%d protocol RFU", val &0x0F); break;
				}

				break;

			default:
				len = sprintf(tbuf, "TA%d 0x%02x", j, val);
				break;
			}

			Tcl_SetVar2(interp, crd, "ATR_VERBOSE", tbuf,
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
		}

		if (tbi[txBits] != N) {		/* TBi follows					*/
			val = resetBuf[i + tbi[txBits]];

			len = sprintf(tbuf, "TB%d", j);
			sprintf(&tbuf[len + 1], "0x%02x", val);
			Tcl_SetVar2(interp, crd, tbuf, &tbuf[len+1],
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
			switch (j) {
			case 1:
				/*
				 * TB1 is interpreted the following way:
				 *	b8 is 0 (or if 1 can be assumed to be 0!)
				 */
				len = sprintf(tbuf, "TB1 0x%02x: programming current is", val);
				switch (val & 0x60) {
				case 0x00:	len += sprintf(&tbuf[len], " 25 mA"); break;
				case 0x20:	len += sprintf(&tbuf[len], " 50 mA"); break;
				case 0x40:	len += sprintf(&tbuf[len], " 100 mA"); break;
				case 0x60:	len += sprintf(&tbuf[len], " RFU mA"); break;
				}
				if ((val & 0x1F) == 0) {
					len += sprintf(&tbuf[len], ", no programming voltage required");
				} else if ((val & 0x1F) >= 5 && (val & 0x1F) <= 25) {
					len += sprintf(&tbuf[len], ", programming voltage %d V", val&0x1F);
				} else {
					len += sprintf(&tbuf[len], ", programming voltage is RFU");
				}
				break;

			case 2:
				/*
				 * TB2 is interpreted the following way:
				 *		50..250 - programming voltage in multiples of 0.1 V
				 *		other   - RFU
				 *
				 * If TB2 exists its value supersedes the TB1 specification
				 * completely.
				 */
				len = sprintf(tbuf, "TB2 0x%02x: programming voltage", val);
				if (val >= 50 && val <= 250)
					len += sprintf(&tbuf[len], " %d.%d Volts", (val / 10) * 10,
							val / 10);
				else
					len += sprintf(&tbuf[len], " RFU");
				len += sprintf(&tbuf[len], " (supercedes TB1)");
				break;

			default:
				len = sprintf(tbuf, "TB%d 0x%02x", j, val);
				break;
			}

			Tcl_SetVar2(interp, crd, "ATR_VERBOSE", tbuf,
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
		}

		if (tci[txBits] != N) {			/* TCi follows				*/
			val = resetBuf[i + tci[txBits]];

			len = sprintf(tbuf, "TC%d", j);
			sprintf(&tbuf[len + 1], "0x%02x", val);
			Tcl_SetVar2(interp, crd, tbuf, &tbuf[len+1],
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
			switch (j) {
			case 1:
				/*
				 * TC1 is interpreted the following way:
				 *		0..254 - extra guard time required (in ETU's)
				 *		255    - guard time reduced from 12 ETU's to 11 ETU's
				 */
				len = sprintf(tbuf, "TC1 0x%02x: extra guard time", val);
				if (val == 255)
					len += sprintf(&tbuf[len], " reduced to 11 ETU's");
				else
					len += sprintf(&tbuf[len], " %d ETU's", val);

				break;

			case 2:
				/*
				 * TC2 is the work waiting time factor. It defines the
				 * maximum time for a response (default: 9600 * D work ETU's)
				 * Defaults to 10, if TC2 is not specified.
				 * NOTE: This could be used to compute the max timeout
				 *       value for init, if the CLK frequency of the
				 *       drive were known.
				 */
				len = sprintf(tbuf, "TC2 0x%02x: work waiting factor %d",
						val, val);
				break;

			default:
				len = sprintf(tbuf, "TC%d 0x%02x", j, val);
				break;
			}

			Tcl_SetVar2(interp, crd, "ATR_VERBOSE", tbuf,
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
		}

		if (tdi[txBits] != N) {				/* TDi follows			*/
			/*
			 * TDi is interpreted the following way:
			 *	High nibble:
			 *		b8 == 1 - TAi follows
			 *		b7 == 1 - TBi follows
			 *		b6 == 1 - TCi follows
			 *		b5 == 1 - TDi follows
			 * Low nibble:
			 *		0 - T0 protocol (asynchronous half duplex character)
			 *		1 - T1 protocol (asynchronous half duplex block)
			 *		2 - RFU (full duplex)
			 *		3 - RFU (full duplex)
			 *		4 - enhanced asynchronous half duplex character
			 *		5..13 - RFU
			 *		14 - Non ISO protocol (e.g.: GSM card)
			 *		15 - reserved for future extensions
			 *
			 */
			val = resetBuf[i + tdi[txBits]];

			len = sprintf(tbuf, "TD%d", j);
			sprintf(&tbuf[len + 1], "0x%02x", val);
			Tcl_SetVar2(interp, crd, tbuf, &tbuf[len+1],
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

			txChars += bits[(val & 0xF0) >> 4];
			len = sprintf(tbuf, "TD%d 0x%02x:", j, val);
			if ((val & 0xF0) == 0)
				len += sprintf(&tbuf[len], " no additional Tx bytes follow");
			else {
				if (val & 0x10)
					len += sprintf(&tbuf[len], " TA%d", j + 1);
				if (val & 0x20)
					len += sprintf(&tbuf[len], " TB%d", j + 1);
				if (val & 0x40)
					len += sprintf(&tbuf[len], " TC%d", j + 1);
				if (val & 0x80)
					len += sprintf(&tbuf[len], " TD%d", j + 1);
				len += sprintf(&tbuf[len], " follow");
			}

			switch (val & 0xF) {
			case 0:	len += sprintf(&tbuf[len], ", T0 protocol (async half duplex char)"); break;
			case 1:	len += sprintf(&tbuf[len], ", T1 protocol (async half duplex block)"); break;
			case 2:	len += sprintf(&tbuf[len], ", T2 protocol RFU (full duplex)"); break;
			case 3:	len += sprintf(&tbuf[len], ", T3 protocol RFU (full duplex)"); break;
			case 4:	len += sprintf(&tbuf[len], ", T4 protocol (enhanced async half duplex char)"); break;
			case 14:len += sprintf(&tbuf[len], ", T14 protocol (Non ISO)"); break;
			default:len += sprintf(&tbuf[len], ", T%d protocol RFU", val &0x0F); break;
			}

			if (val & 0xF)
				haveTck = TRUE;

			Tcl_SetVar2(interp, crd, "ATR_VERBOSE", tbuf,
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

			/*
			 * Go to next group of interface characters, if any
			 */
			if (val & 0xF0) {
				i += bits[txBits];
				txBits = ((val & 0xF0) >> 4);
				++j;
			} else {
				txBits = 0;
			}
		} else {
			txBits = 0;
		}
	} while (txBits);

	/*
	 * Add historical characters to verbose status
	 */
	if (histChars) {
		j = 2 + txChars;
		len = sprintf(tbuf, "Historical characters: ");
		for (i = 0; i < histChars; ++i)
			len += sprintf(&tbuf[len], "%02x", resetBuf[j + i]);
		Tcl_SetVar2(interp, crd, "ATR_VERBOSE", tbuf,
			TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
		for (i = 0; i < histChars; ++i) {
			sprintf(tbuf, "0x%02x", resetBuf[j + i]);
			Tcl_SetVar2(interp, crd, "HIST", tbuf,
				TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
		}
	}

	/*
	 * Compute the XOR of bytes T0 to TCK, if we have a TCK byte
	 */
	if (haveTck) {
		unsigned char tcksum = 0;

		for (i = 1; i < resetLen; ++i)
			tcksum ^= resetBuf[i];

		if (tcksum)
			sprintf(tbuf, "TCK 0x%02x: does not match 0x%02x",
				resetBuf[resetLen - 1], tcksum ^ resetBuf[resetLen - 1]);
		else
			sprintf(tbuf, "TCK 0x%02x: OK", resetBuf[resetLen - 1]);

		Tcl_SetVar2(interp, crd, "ATR_VERBOSE", tbuf,
			TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

		sprintf(tbuf, "0x%02x", resetBuf[resetLen - 1]);				
		Tcl_SetVar2(interp, crd, "TCK", tbuf,
			TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);
	}

	/*
	 * Clear SW1/SW2 in clientData.
	 */
	dp->ack = 0;
	dp->sw1 = dp->sw2 = 0;

	/*
	 * Assemble ATR in return buffer
	 */
	for (i = 0; i < resetLen; ++i) {
		char buf[5];

		sprintf(buf, "0x%02x", (unsigned)resetBuf[i] & 0xFF);
		Tcl_AppendElement(interp, buf);
	}

	Tcl_SetVar2(interp, crd, "ATR", interp->result,
		TCL_GLOBAL_ONLY|TCL_APPEND_VALUE|TCL_LIST_ELEMENT);

	return TCL_OK;
}
#undef N

/*
 * status -- get and display status of last command sent to card
 *		optionally add/delete additional status messages
 */
static int
status(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	card_data *dp = (card_data *)clientData;

	if (argc == 3 && IS_CMD(argv[2], "-verbose")) {
		interp->result = dp->verbose;
		return TCL_OK;
	}
	if (argc == 3 && IS_CMD(argv[2], "-list")) {
		void *state = 0;
		char *code;
		char *meaning;
		char *result;

		while ((code = listCardStatus(&state, &meaning))) {
			char *largv[2];
			largv[0] = code;
			largv[1] = meaning;
			result = Tcl_Merge(2, largv);
			Tcl_AppendElement(interp, result);
		}
		return TCL_OK;
	}
	if (argc == 3 && IS_CMD(argv[2], "-reader")) {
		card_reader *rp = dp->reader;

		if (rp == 0) {
			Tcl_AppendResult(interp, "reader not initialized", NULL);
			return TCL_ERROR;
		}

		Tcl_AppendResult(interp, rp->stat(rp), NULL);
		return TCL_OK;
	}
	if (argc == 3 && IS_CMD(argv[2], "-version")) {
		card_reader *rp = dp->reader;

		if (rp == 0) {
			Tcl_AppendResult(interp, "reader not initialized", NULL);
			return TCL_ERROR;
		}

		Tcl_AppendResult(interp, rp->vers(rp), NULL);
		return TCL_OK;
	}
	if (argc == 4 && IS_CMD(argv[2], "-delete")) {
		int i;

		if (strlen(argv[3]) != 4) {
			Tcl_AppendResult(interp, "bad argument to ", argv[0],
				" ", argv[1], " ", argv[2],
				": must be four hex digits", NULL);
			return TCL_ERROR;
		}
		for (i = 0; i < 4; ++i) {
			if (!isxdigit(argv[3][i]) && argv[3][i] != '?') {
				Tcl_AppendResult(interp, "bad argument to ",
					argv[0], " ", argv[1], " ", argv[2],
					": must be four hex digits",
					NULL);
				return TCL_ERROR;
			}
		}

		if (deleteCardStatus(argv[3]) == 0) {
			Tcl_AppendResult(interp, argv[0], " ", argv[1],
				" ", argv[2],
				": status entry not found",
				NULL);
			return TCL_ERROR;
		}
		return TCL_OK;
	}
	if (argc == 5 && IS_CMD(argv[2], "-add")) {
		int i;

		if (strlen(argv[3]) != 4) {
			Tcl_AppendResult(interp, "bad argument to ", argv[0],
				" ", argv[1], " ", argv[2],
				": must be four hex digits", NULL);
			return TCL_ERROR;
		}
		for (i = 0; i < 4; ++i) {
			if (!isxdigit(argv[3][i]) && argv[3][i] != '?') {
				Tcl_AppendResult(interp, "bad argument to ",
					argv[0], " ", argv[1], " ", argv[2],
					": must be four hex digits",
					NULL);
				return TCL_ERROR;
			}
		}

		if (addCardStatus(argv[3], argv[4]) == 0) {
			Tcl_AppendResult(interp, argv[0], " ", argv[1],
				" ", argv[2],
				": duplicate entry or out of memory",
				NULL);
			return TCL_ERROR;
		}
		return TCL_OK;
	}
	if (argc == 2) {
		char tbuf[256];

		sprintf(tbuf, "0x%02x", dp->sw1);
		Tcl_AppendElement(interp, tbuf);

		sprintf(tbuf, "0x%02x", dp->sw2);
		Tcl_AppendElement(interp, tbuf);

		sprintf(tbuf, "0x%02x", dp->ack);
		Tcl_AppendElement(interp, tbuf);

		return TCL_OK;
	}
	Tcl_AppendResult(interp, "usage is: ", argv[0], " ", argv[1],
		" [-add code text|-delete code|-verbose][-reader][-list]",
		NULL);
	return TCL_ERROR;
}

/*
 * command -- execute a card command
 *
 *	Note that there is a destinction between the TPDU (transmission
 *	protocol data unit) and the APDU (application protocol data unit)
 *	in that the APDU may only have a 4 byte header, which is then
 *	mapped to an appropriate header according to the protocol used (e.g.
 *	If T=0 is used a fifth byte is always available, and T=1 maps
 *	APDU's directly into TPDU blocks).
 *
 *	This currently only supports T=0 protocol.
 *
 *	The send and receive commands are TPDU type calls, the interface
 *	this code provides is an APDU type call. We perform the mapping of
 *	APDU/TPDU and back here and also handle retransmits for card
 *	that provide this information.
 */
static int
command(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	card_data *dp = (card_data *)clientData;
	card_reader *rp = dp->reader;
	int cla;
	int ins;
	int p1;
	int p2;
	int len = 0;
	int largc = 0;
	char **largv = 0;
	char *data;
	int i;
	int j;
	unsigned char tbuf[256];
	unsigned char dbuf[256];
	int status;
	int havesend = 0;
	int haverecv = 0;

	/*
	 * Check for initialized reader
	 */
	if (rp == 0) {
		Tcl_AppendResult(interp, "reader not initialized", NULL);
		return TCL_ERROR;
	}

	/*
	 * Save command in variable
	 */
	data = Tcl_Merge(argc, argv);
	Tcl_SetVar2(interp, crd, "COMMAND", data, TCL_GLOBAL_ONLY);
	Tcl_Free(data);
	data = 0;

	/*
	 * Check parameters
	 */
	if (argc < 8 || argc > 10) {
		Tcl_AppendResult(interp,
			"usage is: card command CLA INS P1 P2 [-send data_list] [-receive length]",
			 NULL);
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[2], &cla) != TCL_OK) {
		Tcl_AppendResult(interp, ": CLA must be numeric", NULL);
		return TCL_ERROR;
	}
	if (cla < 0 || cla > 255) {
		Tcl_AppendResult(interp, "CLA must be in the range 0..255", NULL);
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[3], &ins) != TCL_OK) {
		Tcl_AppendResult(interp, ": INS must be numeric", NULL);
		return TCL_ERROR;
	}
	if (ins & 01) {
		Tcl_AppendResult(interp, "INS must be even", NULL);
		return TCL_ERROR;
	}
	if ((ins & 0xF0) == 0x60 || (ins & 0xF0) == 0x90) {
		Tcl_AppendResult(interp, "INS may not be 0x6? or 0x9?", NULL);
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[4], &p1) != TCL_OK) {
		Tcl_AppendResult(interp, ": P1 must be numeric", NULL);
		return TCL_ERROR;
	}
	if (p1 < 0 || p1 > 255) {
		Tcl_AppendResult(interp, "P1 must be in the range 0..255", NULL);
		return TCL_ERROR;
	}
	if (Tcl_GetInt(interp, argv[5], &p2) != TCL_OK) {
		Tcl_AppendResult(interp, ": P2 must be numeric", NULL);
		return TCL_ERROR;
	}
	if (p2 < 0 || p2 > 255) {
		Tcl_AppendResult(interp, "P2 must be in the range 0..255", NULL);
		return TCL_ERROR;
	}

	/*
	 * Check for send option or receive parameter
	 */
	 for (i = 6; i < argc; ++i) {
		if (strncmp(argv[i], "-send", strlen(argv[i])) == 0) {
			if (++i >= argc) {
				Tcl_AppendResult(interp, "-send requires argument", NULL);
				if (data)
					Tcl_Free(data);
				return TCL_ERROR;
			}

			if (havesend++) {
				Tcl_AppendResult(interp, "duplicate -send option", NULL);
				if (data)
					Tcl_Free(data);
				return TCL_ERROR;
			}

			/*
			 * Handle data hexlist argument
			 */
			if ((status = Tcl_SplitList(interp, argv[i], &largc, &largv)) != TCL_OK) {
				Tcl_AppendResult(interp, "in card command", NULL);
				return status;
			}

			/*
			 * Allocate a data buffer and store converted hexlist there.
			 * The +1 makes sure that we do not call Tcl_Alloc(0).
			 */
			if ((data = Tcl_Alloc(largc + 1)) == NULL) {
				Tcl_AppendResult(interp, "out of memory", NULL);
				return TCL_ERROR;
			}
			for (j = 0; j < largc; ++j) {
				int val;

				if (Tcl_GetInt(interp, largv[j], &val) != TCL_OK ||
					val < 0 || val > 255) {
					Tcl_AppendResult(interp, ": invalid send data", NULL);
					Tcl_Free((char *)largv);
					Tcl_Free(data);
					return TCL_ERROR;
				}
				data[j] = val;
			}
			Tcl_Free((char *)largv);
		} else if (strncmp(argv[i], "-receive", strlen(argv[i])) == 0) {
			if (++i >= argc) {
				Tcl_AppendResult(interp, "-receive requires argument", NULL);
				if (data)
					Tcl_Free(data);
				return TCL_ERROR;
			}

			if (haverecv++) {
				Tcl_AppendResult(interp, "duplicate -receive option", NULL);
				if (data)
					Tcl_Free(data);
				return TCL_ERROR;
			}

			if (Tcl_GetInt(interp, argv[i], &len) != TCL_OK) {
				Tcl_AppendResult(interp, ": -receive argument must be numeric", NULL);
				if (data)
					Tcl_Free(data);
				return TCL_ERROR;
			}
			if (len <= 0) {
				Tcl_AppendResult(interp, "-receive argument must be positive", NULL);
				if (data)
					Tcl_Free(data);
				return TCL_ERROR;
			}
		} else {
			Tcl_AppendResult(interp, "invalid argument: ", argv[i], NULL);
			if (data)
				Tcl_Free(data);
			return TCL_ERROR;
		}
	}

	/*
	 * Reset ACK, SW1, SW2 and verbose message.
	 */
	dp->ack = dp->sw1 = dp->sw2 = 0;
	setStatus(interp, dp);

	/*
	 * All arguments check ok and have been converted to internal form.
	 * Perform translation of APDU into TPDU and call base functions.
	 */
	if ((largc > 0 && largc < 256) || (largc == 0 && len == 0)) {
		/*
		 * ISO7816-4, Annex A.1 and Annex A.3.
		 *		Send a message to the card, either without any data or
		 *		with max. 255 bytes of data.
		 * ISO7816-4, Annex A.4
		 *		Send and receive answer combined into one, send part.
		 */
		status = rp->send(rp, cla, ins, p1, p2, largc, data,
						&dp->ack, &dp->sw1, &dp->sw2);
		setStatus(interp, dp);

		if (data) {
			Tcl_Free(data);
			data = 0;
		}

		if (status != 1) {
			Tcl_AppendResult(interp, rp->stat(rp), NULL);
			return TCL_ERROR;
		}

		if (len == 0) {
			/*
			 * ISO7816-4, Annex A.1 and Annex A.3 end here
			 */
			return TCL_OK;
		}
	}

	if (len > 0 && len < 257 && largc > 0 && largc < 256) {
		/*
		 * ISO7816-4, Annex A.4
		 *		Send and receive answer combined into one
		 *	The send part is identical to the send part in A.1 and
		 *	has already been done in the common code above.
		 *	The receive part is done after the evaluation of the
		 *	status returned by the send part with the ISO defined
		 *	command 'GET RESPONSE' (== CLA 0xC0 0 0 LEN).
		 */
		if ((dp->sw1 & 0x60) == 0x60) {
			if (dp->sw1 != 0x61) {
				/*
				 * Case 4S.1 - command not accepted
				 */
				Tcl_AppendResult(interp, dp->verbose, NULL);
				return TCL_ERROR;
			}

			/*
			 * Case 4S.3 - command accepted with information added.
			 *	Get min(len, dp->sw2) bytes.
			 */
			if (len > dp->sw2)
				len = dp->sw2;

			/*
			 * NOTE: ISO does not specify the CLA?
			 * For now reuse CLA of command. Later should have card
			 * specific value!
			 */
			ins = 0xc0;
			p1 = p2 = 0;

			status = rp->recv(rp, cla, ins, p1, p2, len, dbuf,
						&dp->ack, &dp->sw1, &dp->sw2);
			setStatus(interp, dp);

			if (status != 1) {
				Tcl_AppendResult(interp, rp->stat(rp), NULL);
				return TCL_ERROR;
			}

			for (i = 0; i < len; ++i) {
				sprintf(tbuf, "0x%02x", dbuf[i]);
				Tcl_AppendElement(interp, tbuf);
			}

			return TCL_OK;
		}

		if ((dp->sw1 & 0x90) == 0x90) {
			if (dp->sw1 == 0x90 && dp->sw2 == 0x00) {
				/*
				 * Case 4S.2 - command accepted.
				 *	Get len response bytes.
				 *	Handling is identical to Case 2S.1-4 below
				 *	so we fall through to that code by setting
				 *	largc to 0!
				 */

				/*
				 * NOTE: ISO does not specify the CLA?
				 * For now reuse CLA of command. Later should have card
				 * specific value!
				 */
				ins = 0xc0;
				p1 = p2 = 0x00;
				largc = 0;
			} else {
				/*
				 * Case 4S.4 - copy result unchanged
				 *	Note: As we do not perform the GET RESPONSE there is
				 *		  no data to transfer back to the caller (?implied
				 *		  by ISO standard?).
				 */
				if (data)
					Tcl_Free(data);
				return TCL_OK;
			}
		}

		Tcl_AppendResult(interp, "response not ISO conforming", NULL);
		return TCL_ERROR;
	}

	if (largc == 0 && len > 0 && len < 257) {
		/*
		 * ISO7816-4, Annex A.2
		 *		Receive data of max. 256 bytes from the card.
		 * ISO7816-4, Annex A.4
		 *		Send and receive combined into one, receive part.
		 */
		if (len == 256)			/* Do we need this or will ASE do it? */
			len = 0;

		status = rp->recv(rp, cla, ins, p1, p2, len, dbuf,
					&dp->ack, &dp->sw1, &dp->sw2);
		setStatus(interp, dp);

		if (status != 1) {
			Tcl_AppendResult(interp, rp->stat(rp), NULL);
			return TCL_ERROR;
		}

		/*
		 * Check status bytes
		 */
		if (dp->sw1 == 0x67) {
			/*
			 * Case 2S.2 - len definitly not accepted
			 */
			Tcl_AppendResult(interp, dp->verbose, NULL);
			return TCL_ERROR;
		}

		if (dp->sw1 == 0x6C) {
			/*
			 * Case 2S.3 - len not accepted, correct len indicated
			 */
			len = dp->sw2;

			status = rp->recv(rp, cla, ins, p1, p2, len, dbuf,
						&dp->ack, &dp->sw1, &dp->sw2);
			setStatus(interp, dp);

			if (status != 1) {
				Tcl_AppendResult(interp, rp->stat(rp), NULL);
				return TCL_ERROR;
			}

			/*
			 * Note: ISO7816-4 specifies that if the correct length is
			 *       greater than the expected length the result should
			 *       only be of the length originally given. As we have
			 *       dynamic memory allocation we return all of the info
			 *       received from the card, even if this is bigger than
			 *       expected.
			 */
		}

		/*
		 * ISO 7816 specifies that sw1sw2 codes 0x9??? should copy what we
		 * got unchanged to the caller and that 0x9000 indicates complete
		 * data transfer and that 0x61?? indicates that there is more data.
		 *
		 * We handle all of these by just returning what we have (i.e. len
		 * bytes, a status and ack).
		 */
		for (i = 0; i < len; ++i) {
			sprintf(tbuf, "0x%02x", dbuf[i]);
			Tcl_AppendElement(interp, tbuf);
		}

		return TCL_OK;
	}

	/*
	 * FIX: A.5, A.6, A.7 not yet implemented (long forms)
	 */
	if (data)
		Tcl_Free(data);

	Tcl_AppendResult(interp, "long form not implemented", NULL);
	return TCL_ERROR;
}

/*
 * is -- set type of card
 */
static int
is(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	char *name;

	if (argc == 2) {
		if ((name = Tcl_GetVar2(interp, crd, "CARD", TCL_GLOBAL_ONLY))) {
			Tcl_AppendResult(interp, name, NULL);
			return TCL_OK;
		}
		Tcl_AppendResult(interp, "No card type specified", NULL);
		return TCL_ERROR;
	}

	if (argc == 3) {
		char filename[CARD_PMAX];
		Tcl_SetVar2(interp, crd, "CARD", argv[2], TCL_GLOBAL_ONLY);

		/*
		 * Fix: Perform search in path (Tcl_SplitPath/Tcl_JoinPath)
		 */
		sprintf(filename, "%s.tcl", argv[2]);
		return Tcl_EvalFile(interp, filename);
	}

	Tcl_AppendResult(interp, "usage is: card is [name]", NULL);
	return TCL_ERROR;
}


/*
 * off -- turn card off
 */
static int
off(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	card_data *dp = (card_data *)clientData;
	card_reader *rp = dp->reader;
	int status;

	/*
	 * Check for initialized reader
	 */
	if (rp == 0) {
		Tcl_AppendResult(interp, "card off: reader not initialized",
			NULL);
		return TCL_ERROR;
	}

	if (argc == 2) {
		if ((status =  rp->off(rp)) != 1) {
			sprintf(interp->result, "card off: %s",
				rp->stat(rp));
			return TCL_ERROR;
		}
		dp->verbose[0] = 0;
		dp->sw1 = dp->sw2 = 0;
		dp->ack = 0;
		return TCL_OK;
	}

	Tcl_AppendResult(interp, "usage is: ", argv[0], " off", NULL);
	return TCL_ERROR;
}

/*
 * Central function for card command
 */
static int
card(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			" init|reset|command|status|off [options]\"", NULL);
		return TCL_ERROR;
	}

	/*
	 * Determine command and execute it.
	 */
	if (argv[1][0] == 'i' && IS_CMD(argv[1], "init"))
		return init(clientData, interp, argc, argv);
	else if (argv[1][0] == 'i' && IS_CMD(argv[1], "is"))
		return is(clientData, interp, argc, argv);
	else if (argv[1][0] == 'r' && IS_CMD(argv[1], "reset"))
		return reset(clientData, interp, argc, argv);
	else if (argv[1][0] == 's' && IS_CMD(argv[1], "status"))
		return status(clientData, interp, argc, argv);
	else if (argv[1][0] == 'c' && IS_CMD(argv[1], "command"))
		return command(clientData, interp, argc, argv);
	else if (argv[1][0] == 'o' && IS_CMD(argv[1], "off"))
		return off(clientData, interp, argc, argv);
	else {
		char *card;

		if ((card = Tcl_GetVar2(interp, crd, "CARD", TCL_GLOBAL_ONLY))) {
			Tcl_CmdInfo unused;
			char cmd[256];

			sprintf(cmd, "%s:%s", card, argv[1]);

			if (Tcl_GetCommandInfo(interp, cmd, &unused)) {
				char *cmdp;
				char *tmp = argv[1];
				int status;

				argv[1] = cmd;
				cmdp = Tcl_Merge(argc - 1, argv + 1);
				argv[1] = tmp;
				status = Tcl_Eval(interp, cmdp);
				Tcl_Free(cmdp);
				return status;
			}
		}
		Tcl_AppendResult(interp, "unknown card command \"", argv[1],
			"\"", NULL);
	}
	return TCL_ERROR;
}

/*
 * Function to free allocated space of Card_Init
 */
static void
cardfree(clientData)
ClientData clientData;
{
	card_data *dp = (card_data *)clientData;

	Tcl_Free(dp->verbose);
	Tcl_Free((char *)dp);
}

/*
 * TCL package initialization functions
 */
DllExport int
Card_Init(interp)
Tcl_Interp *interp;		/* interpreter receiving the package	*/
{
	card_data *dp;
	int code = Tcl_PkgProvide(interp, crd, CARD_VERSION);

	if (code != TCL_OK)
		return code;

	if ((dp = (card_data *)Tcl_Alloc(sizeof(card_data)))) {
		dp->reader = 0;
		dp->verbose = Tcl_Alloc(1);
		dp->verbose[0] = 0;
		dp->sw1 = dp->sw2 = 0;
		dp->ack = 0;
	} else {
		interp->result = "Not enough memory for private data";
		return TCL_ERROR;
	}

	Tcl_CreateCommand(interp, crd, card, (ClientData)dp,
		(Tcl_CmdDeleteProc *)cardfree);

	Tcl_CreateCommand(interp, "hexstring", hexString, (ClientData)0,
		(Tcl_CmdDeleteProc *)0);

	Tcl_CreateCommand(interp, "hexlist", hexList, (ClientData)0,
		(Tcl_CmdDeleteProc *)0);

	/*
	 * Unset array variable. This leaves us with a clean start
	 */
	Tcl_UnsetVar(interp, crd, TCL_GLOBAL_ONLY);
	Tcl_SetVar2(interp, crd, "VERSION", CARD_VERSION, TCL_GLOBAL_ONLY);

	/*
	 * Fix: Source module specific startup file
	 */
	return TCL_OK;
}

/*
 * Function to set status information
 */
static void
setStatus(interp, dp)
Tcl_Interp *interp;
card_data *dp;
{
	char tbuf[256];

	sprintf(tbuf, "0x%02x", dp->ack);
	Tcl_SetVar2(interp, crd, "ACK", tbuf, TCL_GLOBAL_ONLY);
	sprintf(tbuf, "0x%02x", dp->sw1);
	Tcl_SetVar2(interp, crd, "SW1", tbuf, TCL_GLOBAL_ONLY);
	sprintf(tbuf, "0x%02x", dp->sw2);
	Tcl_SetVar2(interp, crd, "SW2", tbuf, TCL_GLOBAL_ONLY);

	/*
	 * Set verbose status
	 */
	tbuf[0] = 0;
	decodeCardStatus(dp->sw1, dp->sw2, tbuf, sizeof(tbuf));
	Tcl_Free(dp->verbose);
	dp->verbose = Tcl_Alloc(strlen(tbuf) + 1);
	strcpy(dp->verbose, tbuf);
	Tcl_SetVar2(interp, crd, "STATUS", tbuf, TCL_GLOBAL_ONLY);
}
