# tcl library for easy handling of the Schlumberger CryptoFlex Card
#	(aka ASECard CC4)
#
#	The card supports the following file types:
#	DF -- dedicated file
#	TF -- transparent elementary file (also: EF)
#	CY -- Cyclic elementary file
#	LF -- Linear fixed elementary file
#	LV -- Linear variable elementary file
#
#	The DF 3f00 is the root DF.
#
#	TF 0000 is the CHV file
#	TF 0001 is the KeyINT file
#	TF 0011 is the KeyEXT file
#	TF 0012 is the RSAsec file
#	TF 1012 is the RSApub file
#
#	RSA encryption/decryption requires 1024 bit modulus and public
#	key exponent 0x010001
#
# TODO:
#	Support PRO mode?

set cc4(rootdir)	3f00
set cc4(rootkeynum)	0x01
set cc4(rootkey)	{ 0x47 0x46 0x58 0x49 0x32 0x56 0x78 0x40 }

proc hex x {
	format 0x%02x $x
}

# cc4:add_codes -- add specific CrypoFlex error codes to error code table
proc cc4:add_codes {} {
	foreach code {
		{ 6300 "key verify rejected or wrong cryptogram" }
		{ 6702 "invalid length" }
		{ 67?? "incorrect parameter P3" }
	} {
		# Delete code first, ignoring errors
		catch {[card status -delete [lindex $code 0]]}
		card status -add [lindex $code 0] [lindex $code 1]
	}
}

# cc4:verify_key -- present a key for verification
#	Arguments: key number (0..255) and key value (hexlist)
proc cc4:verify_key {keynum key} {
	card command 0xf0 0x2a 0x00 $keynum -send $key
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
}

# cc4:verify_chv -- present cardholder verification
#	Argument: hexlist of length 8
proc cc4:verify_chv {key} {
	if {[llength $key] != 8} {
		return -code error "CHV is a hex list with 8 elements"
	}
	card command 0xC0 0x20 0x00 0x01 -send $key
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
}

# cc4:verify -- present rootkey
proc cc4:verify {} {
	global cc4
	cc4:select 3f00		;# select root directory
	return [cc4:verify_key $cc4(rootkeynum) $cc4(rootkey)]
}

# cc4:change_chv -- change cardholder verification
#	Arguments: old CHV (hexlist of 8 bytes), new CHV (hexlist of 8 bytes)
proc cc4:change_chv {old new} {
	if {[llength $old] != 8} {
		return -code error "old CHV is a hex list with 8 elements"
	}
	if {[llength $new] != 8} {
		return -code error "new CHV is a hex list with 8 elements"
	}
	card command 0xF0 0x24 0x00 0x10 -send [concat $old $new]
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
}

# cc4:create_record -- create a record
#	Arguments: data to store in record
proc cc4:create_record {data} {
	card command 0xC) 0xE2 0x00 0x00 -send $data
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}	
}

# cc4:create_chv -- create the CHV file
#	Arguments: chv_code unblock_chv_code chv_attempts unblock_attempts
#	Conditions:	Directory where CHV shall be created is selected
#				Directory does not already contain a CHV
proc cc4:create_chv {chv_code unblock_code {chv_attempts 3} {unblock_attempts 10}} {
	if {[llength $chv_code] != 8} {
		return -code error "chv_code is a hex list with 8 elements"
	}
	if {[llength $unblock_code] != 8} {
		return -code error "unblock_code is a hex list with 8 elements"
	}
	if {$chv_attempts <= 0 || $chv_attempts > 255} {
		return -code error "chv_attempts must be in the range 1..255"
	}
	if {$unblock_attempts <= 0 || $unblock_attempts > 255} {
		return -code error "unblock_attempts must be in the range 1..255"
	}

	# Create file - ignore error if already there.
	catch { cc4:create_file 0000 TF {REA=NEV UPD=AUT,1 REH=AUT,1 INV=AUT,1} 24 }
	cc4:select 0000
	cc4:update_binary 0 [concat 0xFF 0xFF 0xFF \
		$chv_code $chv_attempts $chv_attempts \
		$unblock_code $unblock_attempts $unblock_attempts]
}

# cc4:unblock_chv -- unblock CHV
#	Arguments: unblock CHV key, new CHV key
proc cc4:unblock_chv {unblock_key new_key} {
	if {[llength $unblock_key] != 8} {
		return -code error "unblock_key is a hex list with 8 elements"
	}
	if {[llength $new_key] != 8} {
		return -code error "new_key is a hex list with 8 elements"
	}
	card command 0xF0 0x2C 0x00 0x01 -data [concat $unblock_key $new_key]
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
}

# cc4:delete_file -- delete a file
#	Argument: file id (hexstring)
proc cc4:delete_file {id} {
	card command 0xF0 0xE4 0x00 0x00 -send [hexlist hexstring $id]
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
}

# select -- select a file or directory
#	Argument: file id (hexstring)
proc cc4:select {file} {
	card command 0xc0 0xa4 0x00 0x00 -send [hexlist hexstring $file]
	if {[lindex [card status] 0] == 0x61} {
		return [cc4:get_response [lindex [card status] 1]]
	}
	return -code error "select: [card status -verbose]"
}

# cc4:get_response -- issue 'GET RESPONSE' command and retrieve data
proc cc4:get_response {count} {
	set data [card command 0xc0 0xc0 00 00 -receive $count]
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
	return $data
}

# cc4:create_file -- create a file on the card
#	create a file of the given if, type and access. Depending on type
#	further arguments are required (size and record size).
proc cc4:create_file {id type acc args} {
	# We always initialize file to all zeros
	set p1 0x00
	# Set number of record (LF/CY only) to 0
	set p2 0x00
	# Set RFU args to 0
	set a1 0x00
	set a2 0x00
	# Set file size to 0
	set a3 0x00
	set a4 0x00
	# set file id to id given
	set a5 0x[string range $id 0 1]
	set a6 0x[string range $id 2 3]
	# set default file type to invalid (real work done below)
	set a7 0x00
	# set all default access conditions to NEV
	set a8 0x0f	;# update only for EF, RFU for DF
	set a9 0xff
	set a10 0xff
	set a11 0xff
	# Always start as valid
	set a12 0x01
	# set length of data following to max
	set a13 0x04
	# set all keys to keynum 0
	set a14 0x00
	set a15 0x00
	set a16 0x00
	# set record length to 0
	set a17 0x00

	# Determine type of file and get appropriate parameters into variables
	switch -- $type {
	DF {	# Dedicated file
		set a7 0x38

		# Dedicated file requires size that is the size of all members!
		# Size calculation includes headers as required by the files!
		if {[llength $args] != 1} {
			return -code error "DF requires one size argument"
		}
		if {$args%4 != 0} {
			return -code error "DF length must be multiple of 4"
		}

		# Determine size and store it
		set x [hexlist hexstring [format "%04x" $args]]
		if {[llength $x] != 2} {
			return -code error "Invalid file size"
		}
		set a3 [lindex $x 0]
		set a4 [lindex $x 1]

		# Fixup argument list length
		set a13 0x03	;# no record size needed

		# Handle access conditions
		foreach a $acc {
			set x [split $a =]
			if {[llength $x] != 2} {
				return -code error "Bad access condition syntax"
			}
			set ac [cc4:type_ac [lindex $x 1]]
			switch -- [lindex $x 0] {
			DIR {
				set a9 [hex [expr ([lindex $ac 0]<<4)|($a9&0xF)]]
				if {[llength $ac] == 2} {
					set a14 [hex [expr ([lindex $ac 1]<<4)|($a14&0x0F)]]
				}
			}
			DEL {
				set a10 [hex [expr ([lindex $ac 0]<<4)|($a10&0xF)]]
				if {[llength $ac] == 2} {
					set a15 [hex [expr ([lindex $ac 1]<<4)|($a15&0x0F)]]
				}
			}
			CRE {
				set a10 [hex [expr ([lindex $ac 0])|($a10&0xF0)]]
				if {[llength $ac] == 2} {
					set a15 [hex [expr ([lindex $ac 1])|($a15&0xF0)]]
				}
			}
			REH {
				set a11 [hex [expr ([lindex $ac 0]<<4)|($a11&0xF)]]
				if {[llength $ac] == 2} {
					set a16 [hex [expr ([lindex $ac 1]<<4)|($a16&0x0F)]]
				}
			}
			INV {
				set a11 [hex [expr ([lindex $ac 0])|($a11&0xF0)]]
				if {[llength $ac] == 2} {
					set a16 [hex [expr ([lindex $ac 1])|($a16&0xF0)]]
				}
			}
			default	{ return -code error "Bad access type [lindex $x 0]. Must be one of DIR|DEL|CRE|REH|INV" }
			}
		}
	}
	EF -
	TF {	# Transparent EF (requires size)
		set a7 0x01
		if {[llength $args] != 1} {
			return -code error "TF requires one size argument"
		}
		if {$args%4 != 0} {
			return -code error "TF length must be multiple of 4"
		}

		# Determine size and store it
		set x [hexlist hexstring [format "%04x" $args]]
		if {[llength $x] != 2} {
			return -code error "Invalid file size"
		}
		set a3 [lindex $x 0]
		set a4 [lindex $x 1]

		# Fixup argument list length
		set a13 0x03	;# no record size needed
	}
	CY {	# Cyclic EF (requires size and record size)
		set p3 0x11
		set a7 0x06
		if {[llength $args] != 2} {
			return -code error "CY requires two size arguments"
		}
		if {([lindex $args 0]*[lindex $args 1])%4 != 0} {
			return -code error "CY length * recsize must be multiple of 4"
		}

		# Not possible to have more than 255 records (i.e. a3 must be 0)
		if {$a3 != 0} {
			return -code error "Cannot have more than 255 records in a LF"
		}

		# Record size must be < 256
		if {[lindex $args 1] > 255 || [lindex $args 1] < 1} {
			return -code error "Invalid record size"
		}
	}
	LF {	# Linear fixed EF (requires size and record size)
		set a7 0x02
		if {[llength $args] != 2} {
			return -code error "LF requires two size arguments"
		}
		if {([lindex $args 0]*[lindex $args 1])%4 != 0} {
			return -code error "LF length * recsize must be multiple of 4"
		}

		# Determine size and store it
		set x [hexlist hexstring [format "%04x" $args]]
		if {[llength $x] != 2} {
			return -code error "Invalid file size"
		}
		set a3 [lindex $x 0]
		set a4 [lindex $x 1]

		# Not possible to have more than 255 records (i.e. a3 must be 0)
		if {$a3 != 0} {
			return -code error "Cannot have more than 255 records in a LF"
		}

		# Record size must be < 256
		if {[lindex $args 1] > 255 || [lindex $args 1] < 1} {
			return -code error "Invalid record size"
		}
	}
	LV {	# Linear variable EF (requires size)
		set a7 0x04
		if {[llength $args] != 1} {
			return -code error "TV requires one size argument"
		}
		if {$args%4 != 0} {
			return -code error "LV length must be multiple of 4"
		}

		# Determine size and store it
		set x [hexlist hexstring [format "%04x" $args]]
		if {[llength $x] != 2} {
			return -code error "Invalid file size"
		}
		set a3 [lindex $x 0]
		set a4 [lindex $x 1]

		# Not possible to have more than 255 records (i.e. a3 must be 0)
		if {$a3 != 0} {
			return -code error "Cannot have more than 255 records in a LF"
		}

		# Fixup argument list length
		set a13 0x03	;# no record size needed
	}
	default	{ return -code error "Bad file type $type. Must be one of DF|TF|CY|LF|LV" }
	}

	# Handle access conditions of non directory files
	if {[string compare $type DF] != 0} {
		# Handle access conditions
		foreach a $acc {
			set x [split $a =]
			if {[llength $x] != 2} {
				switch -- [lindex $x 0] {
				UPDO {	# update only
					set b8 0x3F
				}
				UPDI {	# update and increment
					set b8 0x7F
				}
				UPDD {	# update and decrement
					set b8 0xBF
				}
				DECI {	# decrement and increment
					set b8 0xFF
				}
				default {	return -code error "Bad access condition syntax" }
				}
				continue
			}
			set ac [cc4:type_ac [lindex $x 1]]
			switch -- [lindex $x 0] {
			REA {	# read and seek
				set a9 [hex [expr ([lindex $ac 0]<<4)|($a9&0xF)]]
				if {[llength $ac] == 2} {
					set a14 [hex [expr ([lindex $ac 1]<<4)|($a14&0x0F)]]
				}
			}
			UPD {	# update/decrement
				set a9 [hex [expr ([lindex $ac 0])|($a9&0xF0)]]
				if {[llength $ac] == 2} {
					set a14 [hex [expr ([lindex $ac 1])|($a14&0xF0)]]
				}
			}
			INC {	# increment
				set a10 [hex [expr ([lindex $ac 0]<<4)|($a10&0xF)]]
				if {[llength $ac] == 2} {
					set a15 [hex [expr ([lindex $ac 1]<<4)|($a15&0x0F)]]
				}
			}
			CRE {	# create record
				set a10 [hex [expr ([lindex $ac 0])|($a10&0xF0)]]
				if {[llength $ac] == 2} {
					set a15 [hex [expr ([lindex $ac 1])|($a15&0xF0)]]
				}
			}
			REH {	# rehabilitate
				set a11 [hex [expr ([lindex $ac 0]<<4)|($a11&0xF)]]
				if {[llength $ac] == 2} {
					set a16 [hex [expr ([lindex $ac 1]<<4)|($a16&0x0F)]]
				}
			}
			INV {	# invalidate
				set a11 [hex [expr ([lindex $ac 0])|($a11&0xF0)]]
				if {[llength $ac] == 2} {
					set a16 [hex [expr ([lindex $ac 1])|($a16&0xF0)]]
				}
			}
			default	{ return -code error "Bad access type [lindex $x 0]. Must be one of DIR|DEL|CRE|REH|INV" }
			}
		}
	}

	# Check type/id correspondence
	#	Done after above switch so we don't have to worry about bad
	#	types.
	set DF {0x3f00}					;# must be DF
	set TF {0x0000 0x0001 0x0011 0x0012 0x1012}	;# must be TF's
	set id 0x$id
	if {[string compare $type DF] == 0} {
		foreach i $TF {
			if {$i == $id} {
				return -code error "$id not a valid DF file id"
			}
		}
	} elseif {[string compare $type TF] == 0 || [string compare $type EF] == 0} {
		foreach i $DF {
			if {$i == $id} {
				return -code error "$id not a valid TF file id"
			}
		}
	} else {
		foreach i [concat $DF $TF] {
			if {$i == $id} {
				return -code error "$id not a valid $type file id"
			}
		}
	}

	# Now for the creation act...
	if {$a13 == 3} {
		card command 0xF0 0xE0 $p1 $p2 \
		-send [list $a1 $a2 $a3 $a4 $a5 $a6 $a7 $a8 $a9 $a10 $a11 $a12 $a13 $a14 $a15 $a16 ]
	} else {
		card command 0xF0 0xE0 $p1 $p2  \
		-send [list $a1 $a2 $a3 $a4 $a5 $a6 $a7 $a8 $a9 $a10 $a11 $a12 $a13 $a14 $a15 $a16 $a17 ]
	}

	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
}

# cc4:response_verbose -- convert a response gotten by get_response
#	after a select to human readable form.
# Note: This only handles responses after a select or reset!
proc cc4:response_verbose {response} {
	if {[llength $response] == 20} {
		# Directory file or master file
		if {[lindex $response 11] == 0} {
			puts -nonewline "Invalidated "
		} else {
			puts -nonewline "Valid "
		}
		if {[string compare [hexstring hexlist [lrange $response 4 5]] "3f00"] == 0} {
			puts "Master file [hexstring hexlist [lrange $response 4 5]]:"
		} else {
			puts "Dedicated file [hexstring hexlist [lrange $response 4 5]]:"
		}
		foreach i [cc4:dfac_type [lrange $response 7 10]] {
			puts "      $i"
		}
		puts "[format %5d [lindex $response 14]] DF's"
		puts "[format %5d [lindex $response 15]] EF's"
		if {[lindex $response 16] == 0 } {
			puts "      no CardHolder Verification file in this DF"
		} else {
			puts "      CardHolder Verification file exists (File ID = 0000)"
			if {[lindex $response 18] == 0} {
				puts "      CHV file invalidated"
			} else {
				puts "      CHV file valid"
			}
			if {[lindex $response 19] == 0} {
				puts "      CHV file unblocking key invalidated"
			} else {
				puts "      CHV file unblocking key valid"
			}
		}
		puts "      Size: 24 + 0 = 24 bytes"
		puts "[format %5d 0x[hexstring hexlist [lrange $response 2 3]]] bytes available"
	} elseif {[llength $response] == 15} {
		if {[lindex $response 11] == 0} {
			puts -nonewline "Invalidated "
		} else {
			puts -nonewline "Valid "
		}
		set special 0x[hexstring hexlist [lrange $response 4 5]]
		if {$special == 0x0000} {
			puts -nonewline "CHV "
		} elseif {$special == 0x0001} {
			puts -nonewline "INT Key "
		} elseif {$special == 0x0011} {
			puts -nonewline "EXT Key "
		} elseif {$special == 0x0012} {
			puts -nonewline "RSA Secret Key "
		} elseif {$special == 0x1012} {
			puts -nonewline "RSA Public Key "
		}
		puts "[cc4:file_type [lindex $response 6]]: [hexstring hexlist [lrange $response 4 5]]:"
		foreach i [cc4:efac_type [lrange $response 7 10]] {
			puts "      $i"
		}
		set size 0x[hexstring hexlist [lrange $response 2 3]]
		puts "      Size: 16 + [format %d $size] = [format %d [expr $size + 16]] bytes"
		if {[lindex $response 14] != 0} {
			puts "[format %5d [lindex $response 14]] bytes per record"
		}
	}
}

# cc4:file_type -- convert file type byte to string
#	expects single byte as argument
proc cc4:file_type {type} {
	if {$type == 0x38} {
		return "DF"
	} elseif {$type == 0x01} {
		return "TF"
	} elseif {$type == 0x02} {
		return "LF"
	} elseif {$type == 0x04} {
		return "LV"
	} elseif {$type == 0x05} {
		return "CY"
	}
	return "Unkown file type '$type'"
}

# cc4:ac_type -- convert access condition code to string
#	expects single byte as argument
proc cc4:ac_type {type} {
	if {$type == 0} {
		return "ALWays" 
	} elseif {$type == 0x01} {
		return "CardHolder Verification"
	} elseif {$type == 0x03} {
		return "PROtected by cryptogram"
	} elseif {$type == 0x04} {
		return "AUThenticated"
	} elseif {$type == 0x06} {
		return "CardHolder Verification and PROtected by cryptogram"
	} elseif {$type == 0x08} {
		return "CardHolder Verification and AUThenticated"
	} elseif {$type == 0x0f} {
		return "NEVer"
	}
	return "Unknown access type '$type'"
}

# cc4:type_ac -- convert string to access condition code
#	Syntax: Access-condition[,key]
#	returns list of access code/keynum
proc cc4:type_ac {arg} {
	set x [split $arg ,]
	switch -- [lindex $x 0] {
	ALW	{ return 0x00 }
	CHV	{ return 0x01 }
	PRO	{ return 0x03 }
	AUT	{ if {[llength $x] != 2} {
			return -code error "AUT protection needs key number!"
		  }
		  return [list 0x04 [format 0x%02x [lindex $x 1]]]
		}
	CHVPRO	{ return 6 }
	CHVAUT	{ if {[llength $x] != 2} {
			return -code error "CHVAUT protection needs key number!"
		  }
		  return [list 0x08 [format 0x%02x [lindex $x 1]]]
		}
	NEV	{ return 0x0f }
	}
	return -code error "Invalid access code [lindex $x 0]"
}

# cc4:dfac_type -- convert DF access conditions to list of strings
#	expects list of four bytes as argument
proc cc4:dfac_type {arg} {
	set dir [expr ([lindex $arg 1] >> 4) & 0x0F]
	set rfu [expr [lindex $arg 1] & 0x0F]
	set del [expr ([lindex $arg 2] >> 4) & 0x0F]
	set cre [expr [lindex $arg 2] & 0x0F]
	set reh [expr ([lindex $arg 3] >> 4) & 0x0F]
	set inv [expr [lindex $arg 3] & 0x0F]

	lappend list "DIR: [cc4:ac_type $dir]"
	lappend list "RFU: [cc4:ac_type $rfu]"
	lappend list "DELETE-FILE: [cc4:ac_type $del]"
	lappend list "CREATE-FILE: [cc4:ac_type $cre]"
	lappend list "REHABILITATE: [cc4:ac_type $reh]"
	lappend list "INVALIDATE: [cc4:ac_type $inv]"
	return $list
}

# cc4:dfac_keys -- show DF access keys
#	expects list of three bytes as argument
proc cc4:dfac_keys {arg} {
	set dir [expr ([lindex $arg 0] >> 4) & 0x0F]
	set rfu [expr [lindex $arg 0] & 0x0F]
	set del [expr ([lindex $arg 1] >> 4) & 0x0F]
	set cre [expr [lindex $arg 1] & 0x0F]
	set reh [expr ([lindex $arg 2] >> 4) & 0x0F]
	set inv [expr [lindex $arg 2] & 0x0F]

	lappend list "DIR: Key is $dir"
	lappend list "RFU: Key is $rfu"
	lappend list "DELETE-FILE: Key is $del"
	lappend list "CREATE-FILE: Key is $cre"
	lappend list "REHABILITATE: Key is $reh"
	lappend list "INVALIDATE: Key is $inv"
	return $list
}

# cc4:efac_type -- convert EF access conditions to list of strings
#	expects list of four bytes as argument
proc cc4:efac_type {arg} {
	set rns [expr ([lindex $arg 1] >> 4) & 0x0F]
	set upd [expr [lindex $arg 1] & 0x0F]
	set inc [expr ([lindex $arg 2] >> 4) & 0x0F]
	set cre [expr [lindex $arg 2] & 0x0F]
	set reh [expr ([lindex $arg 3] >> 4) & 0x0F]
	set inv [expr [lindex $arg 3] & 0x0F]
	set mod [expr ([lindex $arg 0] & 0xC0) >> 6]

	lappend list "READ&SEEK: [cc4:ac_type $rns]"
	lappend list "UPDATE/DECREASE: [cc4:ac_type $upd]"
	lappend list "INCREASE: [cc4:ac_type $inc]"
	if {$mod == 0} {lappend list "Restricted to update only"}
	if {$mod == 1} {lappend list "Restricted to update and increase"}
	if {$mod == 2} {lappend list "Restricted to update and decrease"}
	if {$mod == 3} {lappend list "Restricted to decrease and increase"}
	lappend list "CREATE RECORD: [cc4:ac_type $cre]"
	lappend list "REHABILITATE: [cc4:ac_type $reh]"
	lappend list "INVALIDATE: [cc4:ac_type $inv]"
	return $list
}

# cc4:efac_keys -- show EF access keys
#	expects list of three bytes as argument
proc cc4:efac_keys {arg} {
	set rns [expr ([lindex $arg 0] >> 4) & 0x0F]
	set upd [expr [lindex $arg 0] & 0x0F]
	set inc [expr ([lindex $arg 1] >> 4) & 0x0F]
	set cre [expr [lindex $arg 1] & 0x0F]
	set reh [expr ([lindex $arg 2] >> 4) & 0x0F]
	set inv [expr [lindex $arg 2] & 0x0F]

	lappend list "READ&SEEK: Key is $rns"
	lappend list "UPDATE/DECREASE: Key is $upd"
	lappend list "INCREASE: Key is $inc"
	lappend list "CREATE RECORD: Key is $cre"
	lappend list "REHABILITATE: Key is $reh"
	lappend list "INVALIDATE: Key is $inv"
	return $list
}

# cc4:update_binary -- update a TF
#	Arguments: offset data
#	Conditions:	- File is TF
#				- File is selected
#				- Access conditions fulfilled
proc cc4:update_binary {offset data} {
	set offset [format 0x%04x $offset]
	if {[string length $offset] != 6} {
		return -code error "offset is too big"
	}
	set p1 [hex [expr ($offset>>8)&0xFF]]
	set p2 [hex [expr $offset & 0xFF]]
	card command 0xC0 0xD6 $p1 $p2 -send $data
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
}

# cc4:get_challenge -- get a challenge from the card
#	Arguments: none
#	Conditions: none
proc cc4:get_challenge {} {
	set data [card command 0xC0 0x84 0x00 0x00 -receive 8]
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
	return $data
}

# cc4:invalidate -- invalidate an EF
#	Arguments: none
#	Conditions:	- EF selected
#				- AC for invalidate
proc cc4:invalidate {} {
	card command 0xF0 0x04 0x00 0x00 -send {}
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}	
}

# cc4:rehabilitate -- rehabilitate and EF
#	Arguments: none
#	Conditions:	- EF selected
#				- AC for rehabilitate
proc cc4:rehabilitate {} {
	card command 0xF0 0x44 0x00 0x00 -send {}
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}	
}

# cc4:int_aut -- RSA signature
#	Arguments: key number of RSA key, 128 bytes of data
#	Conditions:	- EF(rsapri) exists and is valid
#				- Access conditions for EF(rsapri) are fulfilled
#	Returns: RSA encrypted data
proc cc4:int_aut {keynum data} {
	if {$keynum < 0 || $keynum > 255} {
		return -code error "keynum must be in the range 0..255"
	}
	if {[llength $data] != 128} {
		return -code error "data is a hex list with 128 elements"
	}
	card command 0x94 0x42 0x00 $keynum -send $data
	set status [card status]
	if {[lindex $status 0] != 0x61 || [lindex $status 1] != 0x80} {
		return -code error [card status -verbose]
	}
	return cc4:get_response 0x80
}

# cc4:read_binary -- read data from a TF
#	Arguments: Offset and length of data to read
#	Conditions:	- Access conditions fullfilled
#				- TF selected
proc cc4:read_binary {offset length} {
	set offset [format 0x%04x $offset]
	if {[string length $offset] != 6} {
		return -code error "offset is too big"
	}
	set p1 [hex [expr ($offset>>8)&0xFF]]
	set p2 [hex [expr $offset & 0xFF]]
	set data [card command 0xC0 0xB0 $p1 $p2 -receive $length]
	set status [card status]
	if {[lindex $status 0] != 0x90 || [lindex $status 1] != 0x00} {
		return -code error [card status -verbose]
	}
	return $data
}

# to execute at source time
cc4:add_codes

# TEMPORARY
# Factor P
set P {\
0x6f 0x36 0x32 0x5e 0xa9 0xc0 0xe1 0x0d 0x4f 0x3c 0xc6 0x22 0x0c 0xe1 0x86 0xf0\
0x09 0xa3 0x07 0x76 0x7b 0x0a 0x44 0xf8 0x69 0xf1 0xca 0x8b 0xac 0x97 0xeb 0xf3\
0xcb 0x02 0x89 0x6b 0xa5 0x61 0x80 0xd8 0x20 0x86 0x80 0xa4 0x11 0xae 0xed 0x9e\
0x3d 0x20 0x23 0x67 0xfd 0x4a 0x5f 0x45 0x2a 0x9b 0xc3 0x51 0x00 0xc3 0xe2 0xbc\
}
# Factor Q
set Q {\
0x6b 0x60 0xcf 0x47 0x5b 0xbb 0x4e 0xc8 0x51 0x01 0x4b 0x2e 0x39 0xad 0xd4 0xf2\
0xf4 0x58 0x43 0x78 0xd1 0xac 0x35 0x9f 0xeb 0x71 0x5b 0x87 0x7f 0x7f 0x49 0x42\
0xfd 0x95 0x4b 0x68 0x2b 0xf0 0xc1 0x3c 0x3a 0x75 0xde 0x4f 0x1c 0xa3 0xb8 0x6d\
0x9f 0x62 0x81 0x20 0x78 0x19 0xd3 0xec 0x4e 0x9f 0x3e 0x80 0x08 0x16 0xfa 0xb8\
}
# a=1/Q mod P
set a {\
0x2a 0x32 0x26 0xa4 0x73 0x12 0x72 0xb2 0x50 0x7d 0x09 0xe1 0xd6 0xe5 0x0c 0xf6\
0x22 0x39 0x04 0x6f 0xfd 0xe8 0xa3 0x0a 0xdf 0xec 0xc2 0x53 0xc3 0x89 0xbd 0x38\
0xd5 0xdb 0xe5 0xe4 0x17 0x1a 0x81 0x7e 0xef 0x29 0xae 0xf8 0xb1 0x91 0xf9 0xa4\
0xea 0x50 0xbb 0xd0 0x9b 0x03 0x3e 0x47 0xec 0x96 0x46 0x30 0xdb 0xe6 0x5a 0x99\
}
# c=Ksec mod (P-1)
set c {\
0x53 0x78 0x82 0x50 0x39 0x42 0x74 0x8e 0x25 0x88 0x87 0x56 0x48 0x5f 0x74 0xbf\
0xb3 0xbc 0xe6 0x0b 0x0c 0x7e 0x9c 0xc3 0xb2 0x88 0x9e 0x16 0xdf 0x0e 0xa1 0xd2\
0x20 0x0a 0xe0 0x2b 0xbb 0x1b 0xaa 0xfb 0xed 0xf8 0x61 0xde 0xe8 0xf6 0x25 0xce\
0xa6 0x32 0x9e 0x3f 0x35 0x7c 0x4a 0x7f 0x74 0x94 0x57 0x9f 0x5d 0x98 0xb2 0x3d\
}
# f=Ksec mod (Q-1)
set f {\
0xb3 0x59 0x6a 0x12 0x53 0x58 0x26 0x9c 0x42 0xbe 0xeb 0x15 0xce 0x84 0xff 0x8d\
0x52 0xd1 0x51 0x03 0x84 0x08 0x95 0x04 0xfe 0x2a 0x9f 0x0c 0x89 0xf8 0x0a 0xe5\
0xc5 0x4f 0xfb 0x81 0xd6 0x92 0xbb 0x0f 0x5f 0x5f 0xd0 0xbf 0x7c 0x77 0xf7 0xf5\
0x8a 0xf6 0x0e 0x82 0xcf 0x2d 0x34 0xc5 0x17 0x66 0xd4 0x32 0x37 0x04 0x1f 0x4a\
}

# pad128 -- pad list to 128 bytes
proc pad128 {l} {
	set length [llength $l]
	if {$length > 128} {
		return -code error "list too long"
	}
	if {$length == 128} {
		return $l
	}
	while {$length < 128} {
		lappend l 0x00
		set length [expr $length + 1]
	}
	return $l
}
