/*
 * gprread.c -- Interface to the GPR400 reader
 *
 * COPYRIGHT:
 *	Copyright 1997 Wolf Geldmacher, Comunicon AG, CH-9052 Niederteufen.
 *	All rights reserved. See file COPYING.
 *
 */

#include "gprread.h"
RCSID "$Id: gprread.c,v 1.3 1998/04/24 18:31:45 wjg Exp $";

#if INCLUDE_GPR != 0

#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include "gpr400.h"

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Create Prototypes for ANSI-C
 */
#ifdef __STDC__
#define P(x) x
#else
#define P(x) ()
#endif

/*
 * Declaration of local functions
 */
static int gpr_init P((
	card_reader *reader,
	int port,
	int timeout));
static int gpr_on P((
	card_reader *reader,
	unsigned char *atr,
	unsigned int *atrlenp));
static int gpr_send P((
	card_reader *reader,
	unsigned char cla,
	unsigned char ins,
	unsigned char p1,
	unsigned char p2,
	unsigned short len,
	unsigned char *data,
	unsigned char *ackp,
	unsigned char *sw1p,
	unsigned char *sw2p));
static int gpr_recv P((
	card_reader *reader,
	unsigned char cla,
	unsigned char ins,
	unsigned char p1,
	unsigned char p2,
	unsigned short len,
	unsigned char *data,
	unsigned char *ackp,
	unsigned char *sw1p,
	unsigned char *sw2p));
static int gpr_off P((
	card_reader *reader));
static char *gpr_msg P((
	card_reader *reader));
static char *gpr_version P((
	card_reader *reader));

#undef P

#ifdef __cplusplus
}
#endif

/*
 * Exported interface to GPR400
 */
card_reader GprReader = {
	"GPR400",		/* reader name				*/
	gpr_init,		/* reader initialize command		*/
	gpr_on,			/* reader on/reset command		*/
	gpr_send,		/* reader send command			*/
	gpr_recv,		/* reader recv command			*/
	gpr_off,		/* reader off command			*/
	gpr_msg,		/* reader status message		*/
	gpr_version,		/* return version information		*/
	0			/* reader status			*/
};

static int gpr_fd = -1;		/* GPR reader file descriptor		*/

static int
gpr_init(reader, port, timeout)
card_reader *reader;
int port;
int timeout;
{
	/*
	 * Initialize the GPR400.
	 *	This is done by opening the device.
	 */
	if (gpr_fd < 0) {
		gpr_fd = open("/dev/gpr400", O_RDWR);
		if (gpr_fd < 0) {
			reader->status = -errno;
			return 0;
		}
	}
	return 1;
}

/*
 * gpr_on -- initialize and reset card
 */
static int
gpr_on(reader, atr, atrlenp)
card_reader *reader;
unsigned char *atr;
unsigned int *atrlenp;
{
	struct gpr400_atr anatr;

	assert(atr && atrlenp && *atrlenp >= 2);
	memset(atr, '\0', *atrlenp);

	if (gpr_fd < 0) {
		return 0;
	}
	if (ioctl(gpr_fd, GPR400_OPEN, &anatr) < 0) {
		reader->status = -errno;
		return 0;
	}
	if (anatr.len > *atrlenp) {
		reader->status = 0xFE;
		return 0;
	}
	*atrlenp = anatr.len;
	memcpy(atr, anatr.data, anatr.len);
	reader->status = anatr.status;
	return anatr.status == 0;
}

static int
gpr_send(reader, cla, ins, p1, p2, len, data, ackp, sw1p, sw2p)
card_reader *reader;	/* Pointer to reader information	*/
unsigned char cla;	/* I: Class byte			*/
unsigned char ins;	/* I: Instruction byte			*/
unsigned char p1;	/* I: P1 byte				*/
unsigned char p2;	/* I: P2 byte				*/
unsigned short len;	/* I: LEN of data			*/
unsigned char *data;	/* I: data to send			*/
unsigned char *ackp;	/* O: Acknowledge byte 			*/
unsigned char *sw1p;	/* O: Contains SW1 for last call	*/
unsigned char *sw2p;	/* O: Contains SW2 for last call	*/
{
	struct gpr400_cmd cmd;

	assert(reader);
	if (gpr_fd < 0) {
		return 0;
	}

	cmd.dir = 0x00;
	cmd.cla = cla;
	cmd.ins = ins;
	cmd.p1 = p1;
	cmd.p2 = p2;
	cmd.len = (unsigned char)len;
	memcpy(cmd.data, data, cmd.len);
	if (ioctl(gpr_fd, GPR400_CMD, &cmd) < 0) {
		reader->status = -errno;
		return 0;
	}
	reader->status = cmd.status;
	*sw1p = cmd.sw1;
	*sw2p = cmd.sw2;
	*ackp = 0;

	return (reader->status == 0);
}

static int
gpr_recv(reader, cla, ins, p1, p2, len, data, ackp, sw1p, sw2p)
card_reader *reader;	/* Pointer to reader information	*/
unsigned char cla;	/* I: Class byte			*/
unsigned char ins;	/* I: Instruction byte			*/
unsigned char p1;	/* I: P1 byte				*/
unsigned char p2;	/* I: P2 byte				*/
unsigned short len;	/* I: LEN of data			*/
unsigned char *data;	/* O: data received			*/
unsigned char *ackp;	/* O: Acknowledge byte 			*/
unsigned char *sw1p;	/* O: Contains SW1 for last call	*/
unsigned char *sw2p;	/* O: Contains SW2 for last call	*/
{
	struct gpr400_cmd cmd;

	assert(reader);
	if (gpr_fd < 0) {
		return 0;
	}

	cmd.dir = 0x01;
	cmd.cla = cla;
	cmd.ins = ins;
	cmd.p1 = p1;
	cmd.p2 = p2;
	cmd.len = (unsigned char)len;
	if (ioctl(gpr_fd, GPR400_CMD, &cmd) < 0) {
		reader->status = -errno;
		return 0;
	}
	memcpy(data, cmd.data, cmd.len);
	reader->status = cmd.status;
	*sw1p = cmd.sw1;
	*sw2p = cmd.sw2;
	*ackp = 0;

	return (reader->status == 0);
}

static int
gpr_off(reader)
card_reader *reader;
{
	assert(reader);

	if (gpr_fd < 0) {
		reader->status = 0xFF;
		return 0;
	}

	close(gpr_fd);
	gpr_fd = -1;

	return 1;
}

/*
 * Function to get GPR400 drive and software information
 */
static char *
gpr_version(reader)
card_reader *reader;
{
	static char tbuf[CARD_BUFSIZE];

	sprintf(tbuf, "GPR400 V%d.%d, Driver V%d.%d", 0, 9, 0, 9);
	return tbuf;
}

/*
 * Function to decode an GPR400 drive status code
 */
static char *
gpr_msg(reader)
card_reader *reader;
{
	static char tbuf[80];

	assert(reader);

	switch (reader->status) {
	case 0x00:	return "Command executed sucessfully";
	case 0x04:	return "Reader command not recognized";
	case 0x05:	return "Smart card driver unknown";
	case 0x10:	return "Card reset response error";
	case 0x12:	return "Invalid (L)ength parameter";
	case 0x14:	return "Card type unknown";
	case 0x15:	return "Card turned off";
	case 0x16:	return "Unsupported V parameter";
	case 0x19:	return "No flash memory present";
	case 0x1a:	return "Error in ISO command format";
	case 0x1b:	return "Incorrect number of parameters";
	case 0x1d:	return "TCK byte in ATR incorrect";
	case 0x1e:	return "Memory access is incorrect";
	case 0x1f:	return "Cannot write flash memory";
	case 0x20:	return "Bad TLV command type <T>";
	case 0xA0:	return "Smart card malfunction (TA1)";
	case 0xA2:	return "Smart card malfunction";
	case 0xA3:	return "Parity error";
	case 0xE4:	return "Invalid procedure byte received";
	case 0xE5:	return "Command with no data";
	case 0xE7:	return "SW1/SW2 not equal to 90 00";
	case 0xF7:	return "Card removed during command";
	case 0xF8:	return "Card in short circuit";
	case 0xFB:	return "No card or card removed";
	/* the next are synthetic reader errors for the interface code */
	case 0xFE:	return "Parameter error";
	case 0xFF:	return "Card reader I/O error";
	default:
		if (reader->status < 0)
			sprintf(tbuf, strerror(-reader->status));
		else
			sprintf(tbuf, "Unknown error code 0x%04x (%d)",
				reader->status, reader->status);
		return tbuf;
	}
}

#endif
