/*
 *	hex.c -- implementation of hex functionality for TCL
 *
 *	hexlist string value|hexstring value|tostring value
 *		Convert strings into lists of hexadecimal constants and vice versa.
 *		hexlists are designed to be operable on by standard TCL commands.
 *
 *		This function in its first form is used to convert a TCL string
 *		into a list of hexadecimal constants (the form required by the
 *		card command). Alternativly the second form converts a hexstring
 *		into a list of hexadecimal constants. The third form converts a
 *		hexlist back into a string with non-printable bytes represented by
 *		their TCL hex escapes.
 *
 *		Returns the result of the conversion or an error message.
 *
 *	hexstring [-space n k] string value|hexlist value|tostring value
 *		Converts strings into hexadecimal strings and vice versa. hexstrings
 *		are designed to be portable representations of binary data.
 *
 *		This function in its first form is used to convert a TCL string
 *		into a hexstring. Alternativly the second form converts a hexlist
 *		into a hexstring. The third form converts a hexstring back into a
 *		string with non-printable bytes represented by their TCL hex escapes.
 *
 *		Hexstrings may contain spaces. Spaces can be introduced into hexstrings
 *		by specifying the -space option to the first two forms of the
 *		command. The -space option will insert k spaces after every
 *		n bytes.
 *
 *		Returns the result of the conversion or an error message.
 *
 *	Copyright 1997, Wolf Geldmacher, Comunicon AG, CH-9052 Niederteufen/AR.
 *	All rights reserved. See file COPYING.
 */

#include "hex.h"
#include <ctype.h>
#include <string.h>
#include "card.h"

RCSID "$Id: hex.c,v 1.3 1998/04/18 20:27:40 wjg Exp $";

/*
 * hexList -- generate a list of hexadecimal digits or an ordinary string from
 *			a list of hexedecimal digits.
 *	This function generates a list of hexadecimal digits from a string or
 *	hexString.
 *	Usage is: hexlist string "...." or hexlist hexstring "...." or
 *		hexlist tostring {...}
 */
int
hexList(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	int length;

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			" command\"", (char *)0);
		return TCL_ERROR;
	}

	length = strlen(argv[1]);
	if (argv[1][0] == 's' && strncmp(argv[1], "string", length) == 0) {
		char *cp;

		/*
		 * Create a hex list from a string
		 */
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" string value\"", (char *)0);
			return TCL_ERROR;
		}

		for (cp = argv[2]; cp && *cp; ++cp) {
			char buf[5];
			int val;

			if (*cp == '\\') {
				int incr;

				val = Tcl_Backslash(cp, &incr);
				cp += incr - 1;
			} else {
				val = (unsigned)*cp & 0xFF;
			}
			sprintf(buf, "0x%02x", val);
			Tcl_AppendElement(interp, buf);
		}
	} else if (argv[1][0] == 'h' && strncmp(argv[1], "hexstring", length) == 0) {
		char *cp = argv[2];
	
		/*
		 * Create a hex list from a hex string
		 */
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" hexstring value\"", (char *)0);
			return TCL_ERROR;
		}

		while (*cp) {
			char buf[5];

			/*
			 * hexstring may contain embedded blanks for better readability
			 * we also allow single hex digits to occur.
			 */
			while (*cp == ' ' || *cp == '\t')
				++cp;
			if (*cp) {
				buf[0] = '0';
				buf[1] = 'x';
				if (isxdigit(*cp))
					buf[2] = *cp++;
				else {
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "Illegal character in hexstring",
						(char *)0);
					return TCL_ERROR;
				}
				if (isxdigit(*cp))
					buf[3] = *cp++;
				else if (*cp && !(*cp == ' ' || *cp == '\t')) {
					/*
					 * hexdigit followed by crap
					 */
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "Illegal character in hexstring",
						(char *)0);
					return TCL_ERROR;
				} else {
					/*
					 * Single hex digit
					 */
					buf[3] = buf[2];
					buf[2] = '0';
				}
				buf[4] = 0;
				Tcl_AppendElement(interp, buf);
			}
		}
	} else if (argv[1][0] == 't' && strncmp(argv[1], "tostring", length) == 0) {
		int result;
		int i;
		int largc;
		char **largv;

		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
				" tostring hexlist\"", (char *)0);
			return TCL_ERROR;
		}

		/*
		 * Split list into elements
		 */
		if ((result = Tcl_SplitList(interp, argv[2], &largc, &largv)) != TCL_OK)
			return result;

		/*
		 * Convert elements to bytes
		 */
		for (i = 0; i < largc; ++i) {
			int val;
			char buf[5];

			if ((result = Tcl_GetInt(interp, largv[i], &val)) != TCL_OK) {
				Tcl_Free((char *)largv);
				return result;
			}
			if (val < 0 || val > 255) {
				Tcl_Free((char *)largv);
				Tcl_ResetResult(interp);
				Tcl_AppendResult(interp, "value in list out of range",
					(char *)0);
				return TCL_ERROR;
			}

			/*
			 * Substitute control characters by their hex equivalent
			 */
			if (iscntrl(val))
				sprintf(buf, "\\x%02x", val);
			else {
				buf[0] = (char)val;
				buf[1] = 0;
			}
			Tcl_AppendResult(interp, buf, (char *)0);
		}

		Tcl_Free((char *)largv);
	} else {
		Tcl_AppendResult(interp, "unknown hexlist command: should be \"",
			"string\", \"hexstring\" or \"tostring\"", (char *)0);
		return TCL_ERROR;
	}

	return TCL_OK;
}

/*
 * hexString -- generate a hex string or ordinary string
 *	This function generates a string of hexadecimal digits from a hexList or
 *	a string. When called as hexstring tostring it generates a normal string
 *	from the argument hexstring.
 *
 *	Usage is: hexstring [-space n/k] string "..." or
 *	          hexstring [-space n/k] hexlist {...} or
 *	          hexstring tostring "..."
 */
int
hexString (clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
	int length;
	int n = 0;				/* n spaces					*/
	int k = 0;				/* per k hexdigits			*/

	if (argc < 2) {
		Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			" [-space n k] command argument\"", (char *)0);
		return TCL_ERROR;
	}

	/*
	 * Handle space option, if specified
	 */
	if (argv[1][0] == '-') {
		if (strcmp(argv[1], "-space")) {
			Tcl_AppendResult(interp, "unknown option: \"", argv[1],
				"\" possible option is -space n k", (char *)0);
			return TCL_ERROR;
		}
		if (argc < 4 || Tcl_GetInt(interp, argv[2], &n) != TCL_OK ||
			Tcl_GetInt(interp, argv[3], &k) != TCL_OK) {
			Tcl_AppendResult(interp, "bad argument(s) to -space option",
				(char *)0);
			return TCL_ERROR;
		}
		if (n < 0 || k < 0) {
			Tcl_AppendResult(interp, "n and k must both be positive", (char *)0);
			return TCL_ERROR;
		}
		argc -= 3; argv += 3;
	}

	length = strlen(argv[1]);
	if (argv[1][0] == 's' && strncmp(argv[1], "string", length) == 0) {
		char *cp;

		/*
		 * Create a hexstring from a string
		 */
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"hexstring",
				" [-space n k] string value\"", (char *)0);
			return TCL_ERROR;
		}

		for (cp = argv[2]; cp && *cp; ++cp) {
			int val;
			char buf[3];

			if (*cp == '\\') {
				int incr;

				val = Tcl_Backslash(cp, &incr);
				cp += incr - 1;
			} else {
				val = (unsigned)*cp & 0xFF;
			}

			sprintf(buf, "%02x", val);

			if (n && cp != argv[2] && (cp - argv[2]) % n == 0) {
				int i;

				for (i = 0; i < k; ++i)
					Tcl_AppendResult(interp, " ", (char *)0);
			}
			Tcl_AppendResult(interp, buf, (char *)0);
		}
	} else if (argv[1][0] == 'h' && strncmp(argv[1], "hexlist", length) == 0) {
		int result;
		int i;
		int largc;
		char **largv;

		/*
		 * Create a hexstring from a hex list
		 */
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"hexstring",
				" [-space n k] hexlist value\"", (char *)0);
			return TCL_ERROR;
		}

		/*
		 * Split list into elements
		 */
		if ((result = Tcl_SplitList(interp, argv[2], &largc, &largv)) != TCL_OK)
			return result;

		/*
		 * Convert elements to hex
		 */
		for (i = 0; i < largc; ++i) {
			int val;
			char buf[5];

			if ((result = Tcl_GetInt(interp, largv[i], &val)) != TCL_OK) {
				Tcl_Free((char *)largv);
				return result;
			}
			if (val < 0 || val > 255) {
				Tcl_Free((char *)largv);
				Tcl_ResetResult(interp);
				Tcl_AppendResult(interp, "value in list out of range",
					(char *)0);
				return TCL_ERROR;
			}

			if (n && i && (i % n) == 0) {
				int j;

				for (j = 0; j < k; ++j)
					Tcl_AppendResult(interp, " ", (char *)0);
			}
			sprintf(buf, "%02x", (unsigned)val & 0xFF);
			Tcl_AppendResult(interp, buf, (char *)0);
		}

		Tcl_Free((char *)largv);
	} else if (argv[1][0] == 't' && strncmp(argv[1], "tostring", length) == 0) {
		char *cp = argv[2];
	
		/*
		 * Create a string from a hexstring
		 */
		if (argc != 3) {
			Tcl_AppendResult(interp, "wrong # args: should be \"hexstring",
				" tostring hexstring\"", (char *)0);
			return TCL_ERROR;
		}

		while (*cp) {
			int val;
			char buf[5];

			/*
			 * hexstring may contain embedded blanks for better readability
			 * we also allow single hex digits to occur.
			 */
			while (*cp == ' ' || *cp == '\t')
				++cp;
			if (*cp) {
				buf[0] = '0';
				buf[1] = 'x';
				if (isxdigit(*cp))
					buf[2] = *cp++;
				else {
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "Illegal character in hexstring",
						(char *)0);
					return TCL_ERROR;
				}
				if (isxdigit(*cp))
					buf[3] = *cp++;
				else if (*cp && !(*cp == ' ' || *cp == '\t')) {
					/*
					 * hexdigit followed by crap
					 */
					Tcl_ResetResult(interp);
					Tcl_AppendResult(interp, "Illegal character in hexstring",
						(char *)0);
					return TCL_ERROR;
				} else {
					/*
					 * Single hex digit
					 */
					buf[3] = buf[2];
					buf[2] = '0';
				}
				buf[4] = 0;
				Tcl_GetInt(interp, buf, &val);

				/*
				 * Substitute control characters by their hex equivalent
				 */
				if (iscntrl(val))
					sprintf(buf, "\\x%02x", val);
				else {
					buf[0] = (char)val;
					buf[1] = 0;
				}
				Tcl_AppendResult(interp, buf, (char *)0);
			}
		}
	} else {
		Tcl_AppendResult(interp, "unknown hexlist command: should be \"",
			"string\", \"hexlist\" or \"tostring\"", (char *)0);
		return TCL_ERROR;
	}

	return TCL_OK;
}
