/*
 * rflxread.c -- Interface to the Schlumberger Reflex 60
 *
 * DESCRIPTION:
 *	This is an interface to the Schlumberger Reflex60 smart card reader.
 *	It currently does not support asynchronous events.
 *
 * COPYRIGHT:
 *	Copyright 1997 Wolf Geldmacher, Comunicon AG, CH-9052 Niederteufen.
 *	All rights reserved. See file COPYING.
 *
 */

#include "rflxread.h"
RCSID "$Id: rflxread.c,v 1.2 1998/04/24 18:31:45 wjg Exp $";

#if INCLUDE_REFLEX != 0

#include <assert.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <termios.h>

#define	OMODE	(O_RDWR|O_NOCTTY)

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Create Prototypes for ANSI-C
 */
#ifdef __STDC__
#define P(x) x
#else
#define P(x) ()
#endif

/*
 * Declaration of local functions
 */
static int reflex_init P((
	card_reader *reader,
	int port,
	int timeout));
static int reflex_on P((
	card_reader *reader,
	unsigned char *atr,
	unsigned int *atrlenp));
static int reflex_send P((
	card_reader *reader,
	unsigned char cla,
	unsigned char ins,
	unsigned char p1,
	unsigned char p2,
	unsigned short len,
	unsigned char *data,
	unsigned char *ackp,
	unsigned char *sw1p,
	unsigned char *sw2p));
static int reflex_recv P((
	card_reader *reader,
	unsigned char cla,
	unsigned char ins,
	unsigned char p1,
	unsigned char p2,
	unsigned short len,
	unsigned char *data,
	unsigned char *ackp,
	unsigned char *sw1p,
	unsigned char *sw2p));
static int reflex_off P((
	card_reader *reader));
static char *reflex_msg P((
	card_reader *reader));
static char *reflex_version P((
	card_reader *reader));
static int t_read P((
	int fd,
	char *buffer,
	int size,
	struct timeval *to));
static void discard P((
	int fd));
static int status P((
	int fd));

#undef P

#ifdef __cplusplus
}
#endif

/*
 * Exported interface to Reflex Reader
 */
card_reader ReflexReader = {
	"Reflex60",		/* reader name				*/
	reflex_init,		/* reader initialize command		*/
	reflex_on,		/* reader on/reset command		*/
	reflex_send,		/* reader send command			*/
	reflex_recv,		/* reader recv command			*/
	reflex_off,		/* reader off command			*/
	reflex_msg,		/* reader status message		*/
	reflex_version,		/* return version information		*/
	0			/* reader status			*/
};

static int reflex_fd = -1;	/* Reflex reader file descriptor	*/
static struct timeval reflex_to;/* Reflex reader timeout		*/
static int reflex_port;		/* Reflex reader port			*/
static struct timeval to_96 = { 0, 10000 };/* 10ms timeout for B9600	*/

static int
reflex_init(reader, port, timeout)
card_reader *reader;
int port;
int timeout;
{
	char buffer[CARD_BUFSIZE];

	/*
	 * If the port is changed we need to close the Reader and
	 * reopen it.
	 */
	if (reflex_fd >= 0 && port != reflex_port) {
		close(reflex_fd);
		reflex_fd = -1;
	}

	/*
	 * Open the Reflex60 Reader
	 */
	if (reflex_fd < 0) {
		struct termios params;

		switch (port) {
		default:
		case 0: reflex_fd = open("/dev/reflex", OMODE);
			break;

		case 1: reflex_fd = open("/dev/cua0", OMODE);
			break;

		case 2: reflex_fd = open("/dev/cua1", OMODE);
			break;

		case 3: reflex_fd = open("/dev/cua2", OMODE);
			break;

		case 4: reflex_fd = open("/dev/cua3", OMODE);
			break;

		case 5: case 6: case 7:
			reader->status = 0xFE;
			break;
		}
		if (reflex_fd < 0) {
			reader->status = -errno;
			return 0;
		}

		/*
		 * Set communication parameters:
		 *	B9600, CS8, EVEN
		 */
		if (tcgetattr(reflex_fd, &params) < 0) {
			reader->status = -errno;
			goto error_return;
		}

		cfsetispeed(&params, B9600);
		cfsetospeed(&params, B9600);

		cfmakeraw(&params);
		params.c_iflag |= INPCK;
		params.c_cflag &= ~PARODD;
		params.c_cflag |= (CREAD|CLOCAL|PARENB);

		if (tcsetattr(reflex_fd, TCSANOW, &params) < 0) {
			reader->status = -errno;
			goto error_return;
		}
	}

	/*
	 * Remember port and timeout
	 */
	reflex_port = port;
	reflex_to.tv_sec = timeout;

	/*
	 * Initialize the reader by performing initial handshake
	 *	write 0x60, read 0x60 and 16 byte reader id
	 */
	buffer[0] = 0x60;
	if (write(reflex_fd, buffer, 1) != 1) {
		reader->status = -errno;
		return 0;
	}
	if (t_read(reflex_fd, buffer, 17, &reflex_to) != 17) {
		reader->status = 0xFF;
		return 0;
	}
	if (buffer[0] != 0x60) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * There might be more information still pending if a card
	 * had been inserted when the reader was initialized.
	 */
	discard(reflex_fd);

	/*
	 * Get reader status
	 */
	reader->status = status(reflex_fd);

	return 1;

error_return:
	close(reflex_fd);
	reflex_fd = -1;
	return 0;
}

/*
 * reflex_on -- initialize and reset card
 *	This assumes that the reset command actually turns the power to
 *	the card on as well. This currently happens but is not documented.
 */
static int
reflex_on(reader, atr, atrlenp)
card_reader *reader;
unsigned char *atr;
unsigned int *atrlenp;
{
	char buffer[CARD_BUFSIZE];
	int len;

	assert(reader);
	assert(atr && atrlenp && *atrlenp >= 2);
	memset(atr, '\0', *atrlenp);

	if (reflex_fd < 0) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Discard pending input, if any
	 */
	discard(reflex_fd);

	/*
	 * Send the reset command
	 */
	buffer[0] = 0x64;
	if (write(reflex_fd, buffer, 1) != 1) {
		reader->status = -errno;
		return 0;
	}

	/*
	 * Get the ACK/NACK byte
	 */
	if (t_read(reflex_fd, buffer, 1, &reflex_to) != 1) {
		reader->status = 0xFF;
		return 0;
	}

	if (buffer[0] != 0x62) {
		reader->status = status(reflex_fd);
		return 0;
	}

	/*
	 * Get ATR. Done in two parts to first get the tag & length and then
	 * the ATR proper.
	 */
	if (t_read(reflex_fd, buffer, 3, &reflex_to) != 3) {
		reader->status = 0xFF;
		return 0;
	}
	if (buffer[0] != 0x64) {
		reader->status = 0xFF;
		return 0;
	}
	len = (((buffer[1] & 0xFF) << 8) | (buffer[2] & 0xFF));
	if (len >= CARD_BUFSIZE || len >= *atrlenp) {
		reader->status = 0xFE;
		return 0;
	}
	if (t_read(reflex_fd, buffer, len, &reflex_to) != len) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Copy ATR and length to caller
	 */
	memcpy(atr, buffer, len);
	*atrlenp = len;

	/*
	 * Set Reader mode.
	 *	ACK handshake following each byte - see reader doc.
	 * FIX: This currently only supports T=0 with default values (i.e. the
	 *	ATR of the card is not evaluated to set things like the extra
	 *	guard time, the work waiting time or error detection codes).
	 */
	buffer[0] = 0x61;
	buffer[1] = 0x00;
	buffer[2] = 0x00;
	buffer[3] = 0x0A;
	buffer[4] = 0x00;
	buffer[5] = 0x00;
	for (len = 0; len < 6; ++len) {
		if (write(reflex_fd, &buffer[len], 1) != 1) {
			reader->status = -errno;
			return 0;
		}
		if (t_read(reflex_fd, &buffer[len], 1, &reflex_to) != 1) {
			reader->status = 0xFF;
			return 0;
		}
		if (buffer[len] != 0x62) {
			reader->status = 0xFF;
			return 0;
		}
	}

	/*
	 * Get and return reader status
	 */
	reader->status = status(reflex_fd);
	return (reader->status == 0);
}

static int
reflex_send(reader, cla, ins, p1, p2, len, data, ackp, sw1p, sw2p)
card_reader *reader;	/* Pointer to reader information	*/
unsigned char cla;	/* I: Class byte			*/
unsigned char ins;	/* I: Instruction byte			*/
unsigned char p1;	/* I: P1 byte				*/
unsigned char p2;	/* I: P2 byte				*/
unsigned short len;	/* I: LEN of data			*/
unsigned char *data;	/* I: data to send			*/
unsigned char *ackp;	/* O: Acknowledge byte 			*/
unsigned char *sw1p;	/* O: Contains SW1 for last call	*/
unsigned char *sw2p;	/* O: Contains SW2 for last call	*/
{
	char buffer[CARD_BUFSIZE];
	int clen;

	assert(reader);
	if (reflex_fd < 0) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Assemble command to reader:
	 *	0x67 rr,ss command
	 *	where rrss is a 16 bit quantity that determines the
	 *	the length of the command.
	 */
	buffer[0] = 0x67;
	clen = len + 5;
	buffer[1] = ((clen >> 8) & 0xFF);
	buffer[2] = (clen & 0xFF);
	buffer[3] = cla;
	buffer[4] = ins;
	buffer[5] = p1;
	buffer[6] = p2;
	buffer[7] = (unsigned char)len;

	if (write(reflex_fd, buffer, 8) != 8) {
		reader->status = -errno;
		return 0;
	}
	if (write(reflex_fd, data, len) != len) {
		reader->status = -errno;
		return 0;
	}

	/*
	 * Fetch ACK byte from reader and check for success.
	 */
	if (t_read(reflex_fd, ackp, 1, &reflex_to) != 1) {
		reader->status = 0xFF;
		return 0;
	}
	if (*ackp != 0x62) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Fetch command result from reader
	 *	This is done in two parts: First get header, then get data.
	 */
	if (t_read(reflex_fd, buffer, 3, &reflex_to) != 3) {
		reader->status = 0xFF;
		return 0;
	}

	if (buffer[0] != 0x64) {
		reader->status = 0xFF;
		return 0;
	}
	clen = (((buffer[1] & 0xFF) << 8) | (buffer[2] & 0xFF));
	if (clen >= CARD_BUFSIZE) {
		reader->status = 0xFF;
		return 0;
	}

	if (t_read(reflex_fd, buffer, clen, &reflex_to) != clen) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Copy status bytes to caller
	 */
	*sw1p = buffer[clen - 2];
	*sw2p = buffer[clen - 1];

	/*
	 * All we expect when sending is the two status bytes. If we got
	 * more than that probably the command should have been a recv
	 * but not a send!
	 */
	if (clen > 2) {
		reader->status = 0xFE;
		return 0;
	}

	/*
	 * Get and return reader status
	 */
	reader->status = status(reflex_fd);
	return (reader->status == 0);
}

static int
reflex_recv(reader, cla, ins, p1, p2, len, data, ackp, sw1p, sw2p)
card_reader *reader;	/* Pointer to reader information	*/
unsigned char cla;	/* I: Class byte			*/
unsigned char ins;	/* I: Instruction byte			*/
unsigned char p1;	/* I: P1 byte				*/
unsigned char p2;	/* I: P2 byte				*/
unsigned short len;	/* I: LEN of data			*/
unsigned char *data;	/* O: data received			*/
unsigned char *ackp;	/* O: Acknowledge byte 			*/
unsigned char *sw1p;	/* O: Contains SW1 for last call	*/
unsigned char *sw2p;	/* O: Contains SW2 for last call	*/
{
	char buffer[CARD_BUFSIZE];
	int clen;

	assert(reader);
	if (reflex_fd < 0) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Assemble command to reader:
	 *	0x67 rr,ss command
	 *	where rrss is a 16 bit quantity that determines the
	 *	the length of the command.
	 */
	buffer[0] = 0x67;
	buffer[1] = 0x00;
	buffer[2] = 0x05;
	buffer[3] = cla;
	buffer[4] = ins;
	buffer[5] = p1;
	buffer[6] = p2;
	buffer[7] = (unsigned char)len;

	if (write(reflex_fd, buffer, 8) != 8) {
		reader->status = -errno;
		return 0;
	}

	/*
	 * Fetch ACK byte from reader and check for success.
	 */
	if (t_read(reflex_fd, ackp, 1, &reflex_to) != 1) {
		reader->status = 0xFF;
		return 0;
	}
	if (*ackp != 0x62) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Fetch command result from reader
	 *	This is done in two parts: First get header, then get data.
	 */
	if (t_read(reflex_fd, buffer, 3, &reflex_to) != 3) {
		reader->status = 0xFF;
		return 0;
	}

	if (buffer[0] != 0x64) {
		reader->status = 0xFF;
		return 0;
	}
	clen = (((buffer[1] & 0xFF) << 8) | (buffer[2] & 0xFF));
	if (clen >= CARD_BUFSIZE) {
		reader->status = 0xFF;
		return 0;
	}

	if (t_read(reflex_fd, buffer, clen, &reflex_to) != clen) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Copy status bytes to caller
	 */
	*sw1p = buffer[clen - 2];
	*sw2p = buffer[clen - 1];

	/*
	 * We expect len bytes plus two status bytes, possibly less.
	 */
	if (clen > len + 2) {
		reader->status = 0xFE;
		return 0;
	}

	/*
	 * Copy buffer to caller
	 */
	memcpy(data, buffer, len);

	/*
	 * Get and return reader status
	 */
	reader->status = status(reflex_fd);
	return (reader->status == 0);
}

static int
reflex_off(reader)
card_reader *reader;
{
	char buffer[CARD_BUFSIZE];

	assert(reader);

	if (reflex_fd < 0) {
		reader->status = 0xFF;
		return 0;
	}

	/*
	 * Send PC_to_RDR_DeActivateRdr, Get Ack, Close fd.
	 */
	buffer[0] = 0x6A;
	if (write(reflex_fd, buffer, 1) != 1) {
		reader->status = -errno;
		return 0;
	}
	if (t_read(reflex_fd, buffer, 1, &reflex_to) != 1) {
		reader->status = -errno;
		return 0;
	}

	/*
	 * Check that reader did ACK
	 */
	if (buffer[0] != 0x62)
		reader->status = 0xFF;
	else
		reader->status = 0;

	close(reflex_fd);
	reflex_fd = -1;

	return 1;
}

/*
 * Function to get Reflex 60 drive and software information
 */
static char *
reflex_version(reader)
card_reader *reader;
{
	static char tbuf[CARD_BUFSIZE];

	sprintf(tbuf, "Reflex60 V%d.%d, Driver V%d.%d", 0, 0, 0, 9);
	return tbuf;
}

/*
 * Function to decode an Reflex 60 drive status code
 */
static char *
reflex_msg(reader)
card_reader *reader;
{
	static char tbuf[80];

	assert(reader);

	/*
	 * All of these error codes are synthetic - the Reflex60 uses a
	 * few bits to signal errors which are converted to byte codes
	 * by the status function below.
	 */
	switch (reader->status) {
	case 0x00:	return "Command executed successfully";
	case 0x04:	return "Reader command not recognized";
	case 0x05:	return "Reader command parameter error";
	case 0x14:	return "Card type unknown";
	case 0x15:	return "Card turned off";

	case 0xA3:	return "Parity error";

	case 0xFB:	return "No card or card removed";
	case 0xFE:	return "Parameter error";
	case 0xFF:	return "Card reader I/O error";
	default:
		if (reader->status < 0)
			sprintf(tbuf, strerror(-reader->status));
		else
			sprintf(tbuf, "Unknown error code 0x%04x (%d)",
				reader->status, reader->status);
		return tbuf;
	}
}

/*
 * t_read -- read with timeout
 *	The timeout is the per byte read timeout (i.e. if we receive a
 *	character shortly before the timeout expires we will wait for the
 *	entire timeout again).
 */
static int
t_read(fd, buffer, count, to)
int fd;
char *buffer;
int count;
struct timeval *to;
{
	fd_set readfds;
	struct timeval tv;
	int rv = 0;
	int rd = 0;

	FD_ZERO(&readfds);
	FD_SET(fd, &readfds);
	tv = *to;

	while (rd < count &&
	       (rv = select(fd + 1, &readfds, NULL, NULL, &tv)) > 0) {

		if ((rv = read(fd, buffer, 1)) != 1)
			break;

		FD_SET(fd, &readfds);
		++buffer;
		++rd;
		tv = *to;
	}

	/*
	 * Propagate error return of select/read if nothing was read
	 */
	if (rd == 0 && rv < 0)
		rd = rv;

	return rd;
}

/*
 * discard -- discard pending input
 *	This function discards input that results from card insertion/
 *	removal events or other spontaneous utterances of the reader.
 */
static void
discard(fd)
int fd;
{
	char buffer[CARD_BUFSIZE];

	while (t_read(fd, buffer, CARD_BUFSIZE, &to_96) > 0)
		continue;
}

/*
 * status -- get reader status
 *	Get the status of the reader. The status consists of four bytes but
 *	only a few bits are used. We compress these bits into a one byte
 *	status code for reflex_msg to decode. Status == 0 designates success.
 *
 *	s1: 0-2 ICC type, 3 ICC present, 4 ICC active
 *	s2: 0 PC_to_RDR_SetMode done, 1 ICC off, 2 Inv. Cmd, 3 Inv. Parameter,
 *	    4 Parity error.
 *	s3: Security Assurance Features Register,  = 0
 *	s4: RFU,  = 0
 */
static int
status(fd)
int fd;
{
	char buffer[CARD_BUFSIZE];

	buffer[0] = 0x65;
	if (write(fd, buffer, 1) != 1)
		return -errno;
	if (t_read(fd, buffer, 5, &reflex_to) != 5)
		return 0xFF;
	if (buffer[0] != 0x61)
		return 0xFF;

	if (!(buffer[2] & 0x01))
		return 0xFB;		/* No card or card removed	*/
	if (buffer[2] & 0x02)
		return 0x15;		/* Card turned off		*/
	if (!(buffer[1] & 0x01))
		return 0x14;		/* Card type unknown		*/
	if (buffer[2] & 0x04)
		return 0x04;		/* Reader cmd not recognized	*/
	if (buffer[2] & 0x08)
		return 0x05;		/* Reader cmd parameter error	*/
	if (buffer[2] & 0x10)
		return 0xA3;		/* parity error			*/

	return 0;
}

#endif
