/*
 * NAME:
 *	status.c -- decoding of card status
 *
 * DESCRIPTION:
 *	The functions in this file are used to decode the standardized
 *	parts of a card's status. The mechanism uses tables that can be
 *	extended by card specific items.
 *
 * COPYRIGHT:
 *	Copyright 1997 Wolf Geldmacher, Comunicon AG, CH-9052 Niederteufen.
 *	All rights reserved. See file COPYING.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "card.h"
#include "status.h"
#include "table.h"
#include "port.h"

RCSID "$Id: status.c,v 1.4 1998/04/18 20:27:41 wjg Exp $";

static struct sw_init {
	char *code;		/* Hex status code...			*/
	char *meaning;		/* ... and associated meaning		*/
} init[] = {			/* ISO7816 predefined stati		*/
	{ "9000", "Normal: No information given" },
	{ "6f00", "Checking error: No precise diagnosis" },
	{ "6e00", "Checking error: Class not supported" },
	{ "6d00", "Checking error: Instruction code not supported or invalid" },
	{ "6c??", "Checking error: Wrong length Le (%03d bytes needed)" },
	{ "6b00", "Checking error: Wrong parameter(s) P1-P2" },
	{ "6a00", "Checking error: No information given" },
	{ "6a80", "Checking error: Incorrect parameters in the data field" },
	{ "6a81", "Checking error: Function not supported" },
	{ "6a82", "Checking error: File not found" },
	{ "6a83", "Checking error: Record not found" },
	{ "6a84", "Checking error: Not enough memory space in the file" },
	{ "6a85", "Checking error: Lc inconsistent with TLV structure" },
	{ "6a86", "Checking error: Incorrect parameters P1-P2" },
	{ "6a87", "Checking error: Lc inconsistent with P1-P2" },
	{ "6a88", "Checking error: Referenced data not found" },
	{ "6900", "Checking error: No information given" },
	{ "6981", "Checking error: Command incompatible with file structure" },
	{ "6982", "Checking error: Security status not satisfied" },
	{ "6983", "Checking error: Authentication method blocked" },
	{ "6984", "Checking error: Referenced data invalidated" },
	{ "6985", "Checking error: Conditions of use not satisfied" },
	{ "6986", "Checking error: Command not allowed (no current EF)" },
	{ "6987", "Checking error: Expected secure messaging data objects missing" },
	{ "6988", "Checking error: Secure messaging data objects incorrect" },
	{ "6800", "Checking error: No information given" },
	{ "6881", "Checking error: Logical channel not supported" },
	{ "6882", "Checking error: Secure messaging not supported" },
	{ "6700", "Checking error: Wrong length" },
	{ "66??", "Execution error: Security-related issue %03d" },
	{ "6500", "Execution error: No information given (state of non-volatile memory changed)" },
	{ "6581", "Execution error: Memory failure (state of non-volatile memory changed)" },
	{ "6400", "Execution error: state of non-volatile memory unchanged" },
	{ "64??", "Execution error: state of non-volatile memory unchanged (RFU = %d)" },
	{ "6300", "Warning: No information given (state of non-volatile memory changed)" },
	{ "6381", "Warning: File filled up by the last write (state of non-volatile memory changed)" },
	{ "63c?", "Warning: Counter provided is %2d (state of non-volatile memory changed)" },
	{ "6200", "Warning: No information given (state of non-volatile memory unchanged)" },
	{ "6281", "Warning: Part of returned data may be corrupted (state of non-volatile memory unchanged)" },
	{ "6282", "Warning: End of file/record reached before reading Le bytes (state of non-volatile memory unchanged)" },
	{ "6283", "Warning: Selected file invalidated (state of non-volatile memory unchanged)" },
	{ "6284", "Warning: File control information not formatted correctly (state of non-volatile memory unchanged)" },
	{ "61??", "Normal: %3d response bytes are still available" },
	{ (char *)0, (char *)0 }
};

static table_t *ISO7816_SW;	/* Decodes ISO7816 status bytes SW1 SW2	*/
static void initCardStatus();	/* Initializes card status table	*/

/*
 * addCardStatus -- add a status message to the status decode table
 *	The sw parameter must contain a four digit hex string. Some of the
 *	nibbles may be coded as '?' and will then be supplied as argument
 *	to the message printed.
 */
int				/* 1 for success, 0 for failure		*/
addCardStatus(sw, message)
const char *sw;			/* Four digit hex code with wildcards	*/
const char *message;		/* and associated message		*/
{
	struct sw_init *swip;
	char nsw[5];		/* lower case hex code */

	assert(sw);
	assert(message);

	if (ISO7816_SW == 0)	/* Initialize, if not already done	*/
		initCardStatus();

	if (strlen(sw) != 4)
		return 0;	/* Invalid entry refused		*/

	/*
	 * Convert code to lower case
	 */
	strcpy(nsw, sw);
	strlwr(nsw);

	if (lookupTable(ISO7816_SW, nsw))
		return 0;	/* duplicate entry refused		*/

	if ((swip = (struct sw_init *)(malloc(sizeof(struct sw_init))))) {
		if ((swip->code = strdup(nsw)) == 0) {
			free(swip);
			return 0;	/* out of memory		*/
		}
		if ((swip->meaning = strdup(message)) == 0) {
			free(swip->code);
			free(swip);
			return 0;	/* out of memory		*/
		}
	}

	if (insertTable(ISO7816_SW, nsw, swip) == 0) {
		free(swip->code);
		free(swip->meaning);
		free(swip);
		return 0;		/* out of memory		*/
	}

	return 1;
}

/*
 * deleteCardStatus -- delete a card status message from the table
 *	The sw parameter must contain a four digit hex string. The
 *	status given must match an entry literally (i.e. including
 *	'?' wildcards) for the entry to be deleted.
 */
int
deleteCardStatus(sw)
const char *sw;
{
	char *meaning;
	char nsw[5];

	if (ISO7816_SW == 0)
		initCardStatus();

	assert(sw);
	if (strlen(sw) != 4)
		return 0;

	/*
	 * Convert code to lower case
	 */
	strcpy(nsw, sw);
	strlwr(nsw);

	if ((meaning = (char *)deleteEntry(ISO7816_SW, nsw))) {
		free(meaning);
		return 1;
	}

	return 0;
}

/*
 * cmp -- comparison function for table lookups
 */
static int
cmp(key, entry)
const void *key;
const void *entry;
{
	return strcmp((char *)key, ((struct sw_init *)entry)->code);
}

/*
 * hash -- status code hash function
 */
static unsigned
hash(key)
const void *key;
{
	unsigned result = 0;
	const char *k = (const char *)key;

	while (*k)
		result = (result << 4) ^ *k++;
	return result;
}

/*
 * initCardStatus -- initialize the state decoding table
 */
static void
initCardStatus()
{
	struct sw_init *swip;

	ISO7816_SW = newTable(S_TSIZE, cmp, hash);
	assert(ISO7816_SW);

	for (swip = init; swip->code; ++swip)
		addCardStatus(swip->code, swip->meaning);
}

/*
 * decodeCardStatus -- decode SW1/SW2 bytes returned from a card
 *	Returns pointer to decoded message (i.e. buffer) or NULL, if
 *	status cannot be decoded. If buffer is NULL malloc is used to
 *	acquire space from the heap and len is ignored. If len is 0
 *	a local static buffer is used to return the result. Otherwise
 *	the user supplied buffer is used. The message must be less than
 *	S_BSIZE bytes in length.
 */
char *
decodeCardStatus(sw1, sw2, buffer, len)
unsigned char sw1;
unsigned char sw2;
char *buffer;
unsigned len;
{
	static char result[S_BSIZE];/* buffer for result			*/
	char key[5];		/* key as string: 4 hex digits + \0	*/
	struct sw_init *swip;	/* pointer to record to locate		*/
	char *message = 0;	/* message to print			*/
	unsigned argument = 0;	/* argument to message to print		*/
	int i;

	if (ISO7816_SW == 0)	/* Initialize, if not already done	*/
		initCardStatus();

	/*
	 * Generate key to find (must generate lower case hex digits!)
	 */
	sprintf(key, "%02x%02x", (unsigned)sw1 & 0xFF, (unsigned)sw2 & 0xFF);
	strcpy(result, key);

	/*
	 * Lookup in table
	 *	First try exact match, then replace digits from right to left
	 *	with wildcards and try again.
	 */
	for (i = 4; i >= 0; --i) {
		if ((swip = lookupTable(ISO7816_SW, key))) {
			message = swip->meaning;
			switch (i) {
			case 4: argument = 0; break;
			case 3: argument = (unsigned)sw2 & 0x0F; break;
			case 2: argument = (unsigned)sw2 & 0xFF; break;
			case 1:	argument = (((unsigned)sw1 & 0x0F) << 8) |
					   ((unsigned)sw2 & 0xFF); break;
			case 0: argument = (((unsigned)sw1 & 0xFF) << 8) |
					   ((unsigned)sw2 & 0xFF); break;
			}
			break;
		} else if (i > 0) {
			key[i - 1] = '?';
		}
	}

	/*
	 * Compute length of result (worst case) and add message to result
	 */
	if (message) {
		if (strlen(message) + 12 > sizeof(result))
			return 0;
		strcat(result, " ");
		i = 6 + sprintf(&result[5], message, argument);
		assert(i < sizeof(result));
	}

	/*
	 * Handle buffer and len argument, possibly copying the static result
	 */
	if (buffer == 0)
		buffer = strdup(result);
	else if (len == 0)
		buffer = result;
	else if (len > strlen(result))
		strcpy(buffer, result);
	else
		buffer = 0;

	return buffer;
}

char *
listCardStatus(state, meaning)
void **state;
char **meaning;
{
	struct sw_init *swip;

	if (ISO7816_SW == 0)	/* Initialize, if not already done	*/
		initCardStatus();

	if ((swip = (struct sw_init *)traverseTable(ISO7816_SW, state))) {
		*meaning = swip->meaning;
		return swip->code;
	}

	return 0;
}
