/*
 * NAME:
 *	table.c -- simple hash table implementation
 *
 * DESCRIPTION:
 *	This is the implementation of the simple hash table functions.
 *
 * COPYRIGHT:
 *	Copyright 1997 Wolf Geldmacher, Comunicon AG, CH-9052 Niederteufen.
 *	All rights reserved. See file COPYING.
 *
 */

#include <stdlib.h>
#include <assert.h>
#include "table.h"
#include "card.h"

RCSID "$Id: table.c,v 1.3 1998/04/18 20:27:41 wjg Exp $";

/*
 * Structure needed for table traversal
 */
typedef struct state state_t;
struct state {
	int idx;
	link_t *lp;
};

/*
 * newLink -- create and initialize a new link
 */
link_t *
newLink(next, data)
link_t *next;
void *data;
{
	link_t *lp;

	if ((lp = (link_t *)malloc(sizeof(link_t)))) {
		lp->next = next;
		lp->data = data;
	}
	return lp;
}

/*
 * deleteLink -- delete a link
 */
void *	
deleteLink(goner)
link_t **goner;
{
	link_t *lp;
	void *data = 0;

	if (goner && (lp = *goner)) {
		*goner = lp->next;
		data = lp->data;
		free(lp);
	}
	return data;
}

/*
 * newTable -- Create/initialize a new table
 */
table_t *
newTable(size, compare, hash)
unsigned size;
cmpf_t compare;
hf_t hash;
{
	table_t *tp;

	assert(size > 0);

	if ((tp = (table_t *)malloc(sizeof(table_t) + (size - 1) * sizeof(link_t *)))) {
		unsigned u;

		tp->size = size;
		tp->compare = compare;
		tp->hash = hash;

		for (u = 0; u < size; ++u)
			tp->base[u] = 0;
	}
	return tp;
}

/*
 * deleteTable -- delete an entire table
 */
void
deleteTable(goner, cleanup)
table_t **goner;
ff_t cleanup;
{
	table_t *tp;

	if (goner && (tp = *goner)) {
		unsigned u;

		for (u = 0; u < tp->size; ++u) {
			link_t *lp = tp->base[u];
			void *data;

			while ((data = deleteLink(&lp)))
				if (cleanup)
					(*cleanup)(data);
		}
		free(tp);
		*goner = 0;
	}
}

/*
 * deleteEntry -- delete an entry in a table
 */
void *
deleteEntry(tp, key)
table_t *tp;
const void *key;
{
	unsigned hval = (*tp->hash)(key) % tp->size;
	link_t **lpp = &tp->base[hval];

	while (*lpp && (*tp->compare)(key, (*lpp)->data))
		lpp = &((*lpp)->next);

	if (*lpp)
		return deleteLink(lpp);
	return 0;
}

/*
 * lookupTable -- lookup a key in a table
 */
void *
lookupTable(tp, key)
table_t *tp;
const void *key;
{
	unsigned hval = (*tp->hash)(key) % tp->size;
	link_t *lp = tp->base[hval];

	while (lp) {
		if ((*tp->compare)(key, lp->data) == 0)
			return lp->data;
		lp = lp->next;
	}

	return 0;
}

/*
 * insertTable -- insert an item into a table
 */
const void *
insertTable(tp, key, data)
table_t *tp;
const void *key;
void *data;
{
	unsigned hval = (*tp->hash)(key) % tp->size;
	link_t *lp = newLink(tp->base[hval], data);

	if (lp) {
		tp->base[hval] = lp;
		return data;
	}
	return 0;
}

/*
 * traverseTable -- traverse a table one item at a time
 */
const void *
traverseTable(tp, vpp)
table_t *tp;
void **vpp;
{
	state_t *sp;
	int i;
	link_t *lp;

	if (vpp == 0)
		return 0;

	if ((sp = (state_t *)*vpp) == 0) {
		sp = (state_t *)malloc(sizeof(state_t));
		if (sp == 0)
			return 0;
		sp->idx = -1;
		sp->lp = 0;
		*vpp = sp;
	}

	lp = sp->lp;
	if (lp && lp->next) {
		sp->lp = lp->next;
		return sp->lp->data;
	}
	for (i = sp->idx + 1; i < (int)tp->size; ++i) {
		if (tp->base[i]) {
			sp->idx = i;
			sp->lp = tp->base[i];
			return sp->lp->data;
		}
	}
	free(sp);
	*vpp = 0;
	return 0;
}
